<?php
/*
Plugin Name: Dessky Security
Plugin URI:  http://dessky.com/plugin/dessky-security
Description: Dessky Security is the ultralight plugin for basic Security Hardening. It is specially designed not to drain any resources from your website.
Author: Dessky
Version: 1.1
Author URI: http://dessky.com
*/


/**
 * Main file to control the plugin.
 *
 * @package   Dessky Security
 * @author    Dessky
 * @copyright Since 2015 Dessky.com
 * @license   Released under the GPL - see LICENSE file for details.
 * @link      http://dessky.com/plugin/dessky-security
 * @since     File available since Release 1.0
 */


/**
 * Plugin dependencies.
 *
 * List of required functions for the execution of this plugin, we are assuming
 * that this site was built on top of the WordPress project, and that it is
 * being loaded through a pluggable system, these functions most be defined
 * before to continue.
 *
 * @var array
 */
$desskyscan_dependencies = array(
    'wp',
    'wp_die',
    'add_action',
    'remove_action',
    'wp_remote_get',
    'wp_remote_post',
);

// Terminate execution if any of the functions mentioned above is not defined.
foreach ( $desskyscan_dependencies as $dependency ) {
    if ( ! function_exists( $dependency ) ) {
        exit(0);
    }
}

/**
 * Plugin's constants.
 *
 * These constants will hold the basic information of the plugin, file/folder
 * paths, version numbers, read-only variables that will affect the functioning
 * of the rest of the code. The conditional will act as a container helping in
 * the readability of the code considering the total number of lines that this
 * file will have.
 */

/**
 * Unique name of the plugin through out all the code.
 */
define( 'DESSKYSCAN', 'desskyscan' );

/**
 * Current version of the plugin's code.
 */
define( 'DESSKYSCAN_VERSION', '1.1' );

/**
 * The name of the Dessky plugin main file.
 */
define( 'DESSKYSCAN_PLUGIN_FILE', 'dessky-security.php' );

/**
 * The name of the folder where the plugin's files will be located.
 */
define( 'DESSKYSCAN_PLUGIN_FOLDER', 'dessky-security' );

/**
 * The fullpath where the plugin's files will be located.
 */
define( 'DESSKYSCAN_PLUGIN_PATH', WP_PLUGIN_DIR.'/'.DESSKYSCAN_PLUGIN_FOLDER );

/**
 * The fullpath of the main plugin file.
 */
define( 'DESSKYSCAN_PLUGIN_FILEPATH', DESSKYSCAN_PLUGIN_PATH.'/'.DESSKYSCAN_PLUGIN_FILE );

/**
 * The local URL where the plugin's files and assets are served.
 */
define( 'DESSKYSCAN_URL', rtrim( plugin_dir_url( DESSKYSCAN_PLUGIN_FILEPATH ), '/' ) );

/**
 * Checksum of this file to check the integrity of the plugin.
 */
define( 'DESSKYSCAN_PLUGIN_CHECKSUM', @md5_file( DESSKYSCAN_PLUGIN_FILEPATH ) );

/**
 * The minimum quantity of seconds to wait before each filesystem scan.
 */
define( 'DESSKYSCAN_MINIMUM_RUNTIME', 10800 );

/**
 * Plugin's global variables.
 *
 * These variables will be defined globally to allow the inclusion in multiple
 * functions and classes defined in the libraries loaded by this plugin. The
 * conditional will act as a container helping in the readability of the code
 * considering the total number of lines that this file will have.
 */
if ( defined( 'DESSKYSCAN' ) ) {
    /**
     * List an associative array with the sub-pages of this plugin.
     *
     * @return array
     */
    $desskyscan_pages = array(
        'desskyscan_hardening' => 'Hardening',
        'desskyscan_infosys' => 'Site Info',
    );

    $desskyscan_verify_ssl_cert = array(
        'true' => 'Verify peer\'s cert',
        'false' => 'Stop peer\'s cert verification',
    );

    $desskyscan_no_notices_in = array(
        /* Value of the page parameter to ignore. */
    );

    /**
     * Remove the WordPress generator meta-tag from the source code.
     */
    remove_action( 'wp_head', 'wp_generator' );
    add_action( 'admin_enqueue_scripts', 'DesskyScanInterface::enqueue_scripts', 1 );
    add_action( 'admin_menu', 'DesskyScanInterface::add_interface_menu' );

    /**
     * Attach Ajax requests to a custom page handler.
     */
    foreach ( $desskyscan_pages as $page_func => $page_title ) {
        $ajax_func = $page_func . '_ajax';

        if ( function_exists( $ajax_func ) ) {
            add_action( 'wp_ajax_' . $ajax_func, $ajax_func );
        }
    }
}

/**
 * Miscellaneous library.
 *
 * Multiple and generic functions that will be used through out the code of
 * other libraries extending from this and functions defined in other files, be
 * aware of the hierarchy and check the other libraries for duplicated methods.
 */
class DesskyScan {

    /**
     * Class constructor.
     */
    public function __construct(){
    }

    /**
     * Return name of a variable with the plugin's prefix (if needed).
     *
     * To facilitate the development, you can prefix the name of the key in the
     * request (when accessing it) with a single colon, this function will
     * automatically replace that character with the unique identifier of the
     * plugin.
     *
     * @param  string $var_name Name of a variable with an optional colon at the beginning.
     * @return string           Full name of the variable with the extra characters (if needed).
     */
    public static function variable_prefix( $var_name = '' ){
        if ( preg_match( '/^:(.*)/', $var_name, $match ) ) {
            $var_name = sprintf( '%s_%s', DESSKYSCAN, $match[1] );
        }

        return $var_name;
    }

    /**
     * Gets the value of a configuration option.
     *
     * @param  string $property The configuration option name.
     * @return string           Value of the configuration option as a string on success.
     */
    public static function ini_get( $property = '' ){
        $ini_value = ini_get( $property );

        if ( empty($ini_value) || is_null( $ini_value ) ) {
            switch ( $property ) {
                case 'error_log': $ini_value = 'error_log'; break;
                case 'safe_mode': $ini_value = 'Off'; break;
                case 'allow_url_fopen': $ini_value = '1'; break;
                case 'memory_limit': $ini_value = '128M'; break;
                case 'upload_max_filesize': $ini_value = '2M'; break;
                case 'post_max_size': $ini_value = '8M'; break;
                case 'max_execution_time': $ini_value = '30'; break;
                case 'max_input_time': $ini_value = '-1'; break;
            }
        }

        if ( $property == 'error_log' ) {
            $ini_value = basename( $ini_value );
        }

        return $ini_value;
    }

    /**
     * Encodes the less-than, greater-than, ampersand, double quote and single quote
     * characters, will never double encode entities.
     *
     * @param  string $text The text which is to be encoded.
     * @return string       The encoded text with HTML entities.
     */
    public static function escape( $text = '' ){
        // Escape the value of the variable using a built-in function if possible.
        if ( function_exists( 'esc_attr' ) ) {
            $text = esc_attr( $text );
        } else {
            $text = htmlspecialchars( $text );
        }

        return $text;
    }

    /**
     * Translate a given number in bytes to a human readable file size using the
     * a approximate value in Kylo, Mega, Giga, etc.
     *
     * @link   http://www.php.net/manual/en/function.filesize.php#106569
     * @param  integer $bytes    An integer representing a file size in bytes.
     * @param  integer $decimals How many decimals should be returned after the translation.
     * @return string            Human readable representation of the given number in Kylo, Mega, Giga, etc.
     */
    public static function human_filesize( $bytes = 0, $decimals = 2 ){
        $sz = 'BKMGTP';
        $factor = floor( (strlen( $bytes ) - 1) / 3 );
        return sprintf( "%.{$decimals}f", $bytes / pow( 1024, $factor ) ) . @$sz[ $factor ];
    }

    /**
     * Check whether the current site is working as a multi-site instance.
     *
     * @return boolean Either TRUE or FALSE in case WordPress is being used as a multi-site instance.
     */
    public static function is_multisite(){
        if (
            function_exists( 'is_multisite' )
            && is_multisite()
        ) {
            return true;
        }

        return false;
    }

    /**
     * Find and retrieve the current version of Wordpress installed.
     *
     * @return string The version number of Wordpress installed.
     */
    public static function site_version(){
        global $wp_version;

        if ( $wp_version === null ) {
            $wp_version_path = ABSPATH . WPINC . '/version.php';

            if ( file_exists( $wp_version_path ) ) {
                include($wp_version_path);
                $wp_version = isset($wp_version) ? $wp_version : '0.0';
            } else {
                $option_version = get_option( 'version' );
                $wp_version = $option_version ? $option_version : '0.0';
            }
        }

        $wp_version = self::escape( $wp_version );

        return $wp_version;
    }

    /**
     * Find and retrieve the absolute path of the WordPress configuration file.
     *
     * @return string Absolute path of the WordPress configuration file.
     */
    public static function get_wpconfig_path(){
        if ( defined( 'ABSPATH' ) ) {
            $file_path = ABSPATH . '/wp-config.php';

            // if wp-config.php doesn't exist, or is not readable check one directory up.
            if ( ! file_exists( $file_path ) ) {
                $file_path = ABSPATH . '/../wp-config.php';
            }

            // Remove duplicated double slashes.
            $file_path = @realpath( $file_path );

            if ( $file_path ) {
                return $file_path;
            }
        }

        return false;
    }

    /**
     * Find and retrieve the absolute path of the main WordPress htaccess file.
     *
     * @return string Absolute path of the main WordPress htaccess file.
     */
    public static function get_htaccess_path(){
        if ( defined( 'ABSPATH' ) ) {
            $base_dirs = array(
                rtrim( ABSPATH, '/' ),
                dirname( ABSPATH ),
                dirname( dirname( ABSPATH ) ),
            );

            foreach ( $base_dirs as $base_dir ) {
                $htaccess_path = sprintf( '%s/.htaccess', $base_dir );

                if ( file_exists( $htaccess_path ) ) {
                    return $htaccess_path;
                }
            }
        }

        return false;
    }

    /**
     * Get the pattern of the definition related with a WordPress secret key.
     *
     * @return string Secret key definition pattern.
     */
    public static function secret_key_pattern(){
        return '/define\(\s*\'([A-Z_]+)\',(\s*)\'(.+)\'\s*\);/';
    }

    /**
     * Retrieve the real ip address of the user in the current request.
     *
     * @param  boolean $return_header Whether the header name where the address was found must be returned.
     * @return string                 The real ip address of the user in the current request.
     */
    public static function get_remote_addr( $return_header = false ){
        $remote_addr = '';
        $header_used = 'unknown';

        if ( isset($_SERVER['REMOTE_ADDR']) ) {
            $remote_addr = $_SERVER['REMOTE_ADDR'];
            $header_used = 'REMOTE_ADDR';
        }

        if ( $remote_addr == '::1' ) {
            $remote_addr = '127.0.0.1';
        }

        if ( $return_header ) {
            return $header_used;
        }

        return $remote_addr;
    }

    /**
     * Return the HTTP header used to retrieve the remote address.
     *
     * @return string The HTTP header used to retrieve the remote address.
     */
    public static function get_remote_addr_header(){
        return self::get_remote_addr( true );
    }

    /**
     * Get the clean version of the current domain.
     *
     * @return string The domain of the current site.
     */
    public static function get_domain( $return_tld = false ){
        if ( function_exists( 'get_site_url' ) ) {
            $site_url = get_site_url();
            $pattern = '/([fhtps]+:\/\/)?([^:\/]+)(:[0-9:]+)?(\/.*)?/';
            $replacement = ( $return_tld === true ) ? '$2' : '$2$3$4';
            $domain_name = @preg_replace( $pattern, $replacement, $site_url );

            return $domain_name;
        }

        return false;
    }

    /**
     * Get top-level domain (TLD) of the website.
     *
     * @return string Top-level domain (TLD) of the website.
     */
    public static function get_top_level_domain(){
        return self::get_domain( true );
    }

    /**
     * Check whether the DNS lookups should be execute or not.
     *
     * DNS lookups are only necessary if you are planning to use a reverse proxy or
     * firewall, this is used to set the correct IP address when
     * the firewall/proxy filters the requests. If you are not planning to use any
     * of these is better to disable this option, otherwise the load time of your
     * site may be affected.
     *
     * @return boolean True if the DNS lookups should be executed, false otherwise.
     */
    public static function execute_dns_lookups(){
        if (
            DesskyScanOption::get_option( ':dns_lookups' ) === 'disabled'
        ) {
            return false;
        }

        return true;
    }

    /**
     * Check whether the site is behind the Proxy.
     *
     * @param  boolean $verbose Return an array with the hostname, address, and status, or not.
     * @return boolean          Either TRUE or FALSE if the site is behind Proxy.
     */
    public static function check_hosting_server( $verbose = false ){
        $http_host = self::get_top_level_domain();

        if ( self::execute_dns_lookups() ) {
            $host_by_addr = @gethostbyname( $http_host );
            $host_by_name = @gethostbyaddr( $host_by_addr );
            $status = false;
        } else {
            $status = false;
            $host_by_addr = '::1';
            $host_by_name = 'localhost';
        }
            $status = true;

        if ( $verbose ) {
            return array(
                'http_host' => $http_host,
                'host_name' => $host_by_name,
                'host_addr' => $host_by_addr,
                'status' => $status,
            );
        }

        return $status;
    }

    /**
     * Retrieve a list of all admin user accounts.
     *
     * @return array List of admin users, false otherwise.
     */
    public static function get_admin_users(){
        if ( function_exists( 'get_users' ) ) {
            $args = array( 'role' => 'administrator' );

            return get_users( $args );
        }

        return false;
    }

    /**
     * Returns the current time measured in the number of seconds since the Unix Epoch.
     *
     * @return integer Return current Unix timestamp.
     */
    public static function local_time(){
        if ( function_exists( 'current_time' ) ) {
            return current_time( 'timestamp' );
        } else {
            return time();
        }
    }

    /**
     * Retrieve the date in localized format, based on timestamp.
     *
     * If the locale specifies the locale month and weekday, then the locale will
     * take over the format for the date. If it isn't, then the date format string
     * will be used instead.
     *
     * @param  integer $timestamp Unix timestamp.
     * @return string             The date, translated if locale specifies it.
     */
    public static function datetime( $timestamp = 0 ){
        if ( is_numeric( $timestamp ) && $timestamp > 0 ) {
            $gmt_offset = get_option( 'gmt_offset' );
            $date_format = get_option( 'date_format' );
            $time_format = get_option( 'time_format' );
            $tz_format = sprintf( '%s %s', $date_format, $time_format );

            if ( is_numeric( $gmt_offset ) ) {
                $timestamp += ( $gmt_offset * 3600 );
            }

            return date_i18n( $tz_format, $timestamp );
        }

        return null;
    }

    /**
     * Return the time passed since the specified timestamp until now.
     *
     * @param  integer $timestamp The Unix time number of the date/time before now.
     * @return string             The time passed since the timestamp specified.
     */
    public static function time_ago( $timestamp = 0 ){
        if ( ! is_numeric( $timestamp ) ) {
            $timestamp = strtotime( $timestamp );
        }

        $local_time = self::local_time();
        $diff = abs( $local_time - intval( $timestamp ) );

        if ( $diff == 0 ) {
            return 'just now';
        }

        $intervals = array(
            1                => array( 'year', 31556926, ),
            $diff < 31556926 => array( 'month', 2592000, ),
            $diff < 2592000  => array( 'week', 604800, ),
            $diff < 604800   => array( 'day', 86400, ),
            $diff < 86400    => array( 'hour', 3600, ),
            $diff < 3600     => array( 'minute', 60, ),
            $diff < 60       => array( 'second', 1, ),
        );

        $value = floor( $diff / $intervals[1][1] );
        $time_ago = sprintf(
            '%s %s%s ago',
            $value,
            $intervals[1][0],
            ( $value > 1 ? 's' : '' )
        );

        return $time_ago;
    }

    /**
     * Check whether a variable contains a serialized data or not.
     *
     * @param  string  $data The data that will be checked.
     * @return boolean       TRUE if the data was serialized, FALSE otherwise.
     */
    public static function is_serialized( $data = '' ){
        return ( is_string( $data ) && preg_match( '/^(a|O):[0-9]+:.+/', $data ) );
    }

    /**
     * Check whether an IP address has a valid format or not.
     *
     * @param  string  $remote_addr The host IP address.
     * @return boolean              Whether the IP address specified is valid or not.
     */
    public static function is_valid_ip( $remote_addr = '' ){
        if ( function_exists( 'filter_var' ) ) {
            return (bool) filter_var( $remote_addr, FILTER_VALIDATE_IP );
        } elseif ( strlen( $remote_addr ) >= 7 ) {
            $pattern = '/^([0-9]{1,3}\.) {3}[0-9]{1,3}$/';

            if ( preg_match( $pattern, $remote_addr, $match ) ) {
                for ( $i = 0; $i < 4; $i++ ) {
                    if ( $match[ $i ] > 255 ) {
                        return false;
                    }
                }

                return true;
            }
        }

        return false;
    }

    /**
     * Check whether a regular expression is valid or not.
     *
     * @param  string  $pattern The regular expression to check.
     * @return boolean          True if the regular expression is valid, false otherwise.
     */
    public static function is_valid_pattern( $pattern = '' ){
        return (bool) (
            is_string( $pattern )
            && ! empty( $pattern )
            && @preg_match( $pattern, null ) !== false
        );
    }

    /**
     * Cut a long text to the length specified, and append suspensive points at the end.
     *
     * @param  string  $text   String of characters that will be cut.
     * @param  integer $length Maximum length of the returned string, default is 10.
     * @return string          Short version of the text specified.
     */
    public static function excerpt( $text = '', $length = 10 ){
        $text_length = strlen( $text );

        if ( $text_length > $length ) {
            return substr( $text, 0, $length ) . '...';
        }

        return $text;
    }

    /**
     * Check whether an list is a multidimensional array or not.
     *
     * @param  array   $list An array or multidimensional array of different values.
     * @return boolean       TRUE if the list is multidimensional, FALSE otherwise.
     */
    public static function is_multi_list( $list = array() ){
        if ( ! empty($list) ) {
            foreach ( (array) $list as $item ) {
                if ( is_array( $item ) ) {
                    return true;
                }
            }
        }

        return false;
    }

    /**
     * Join array elements with a string no matter if it is multidimensional.
     *
     * @param  string $separator Character that will act as a separator, default to an empty string.
     * @param  array  $list      The array of strings to implode.
     * @return string            String of all the items in the list, with the separator between them.
     */
    public static function implode( $separator = '', $list = array() ){
        if ( self::is_multi_list( $list ) ) {
            $pieces = array();

            foreach ( $list as $items ) {
                $pieces[] = @implode( $separator, $items );
            }

            $joined_pieces = '(' . implode( '), (', $pieces ) . ')';

            return $joined_pieces;
        } else {
            return implode( $separator, $list );
        }
    }

    /**
     * Check whether the site is running over the Nginx web server.
     *
     * @return boolean TRUE if the site is running over Nginx, FALSE otherwise.
     */
    public static function is_nginx_server(){
        return (bool) preg_match( '/^nginx(\/[0-9\.]+)?$/', @$_SERVER['SERVER_SOFTWARE'] );
    }

    /**
     * Check whether the site is running over the Nginx web server.
     *
     * @return boolean TRUE if the site is running over Nginx, FALSE otherwise.
     */
    public static function is_iis_server(){
        return (bool) preg_match( '/Microsoft-IIS/i', @$_SERVER['SERVER_SOFTWARE'] );
    }

}

/**
 * System events, reports and actions.
 *
 * An event is an action or occurrence detected by the program that may be
 * handled by the program. Typically events are handled synchronously with the
 * program flow, that is, the program has one or more dedicated places where
 * events are handled, frequently an event loop. Typical sources of events
 * include the user; another source is a hardware device such as a timer. Any
 * program can trigger its own custom set of events as well, e.g. to communicate
 * the completion of a task. A computer program that changes its behavior in
 * response to events is said to be event-driven, often with the goal of being
 * interactive.
 *
 * @see http://en.wikipedia.org/wiki/Event_(computing)
 */
class DesskyScanEvent extends DesskyScan {

    /**
     * Check whether the current WordPress version must be reported to the API
     * service or not, this is to avoid duplicated information in the audit logs.
     *
     * @return boolean TRUE if the current WordPress version must be reported, FALSE otherwise.
     */
    private static function report_site_version(){
        $option_name = ':site_version';
        $reported_version = DesskyScanOption::get_option( $option_name );
        $wp_version = self::site_version();

        if ( $reported_version != $wp_version ) {
            DesskyScanEvent::report_info_event( 'WordPress version detected ' . $wp_version );
            DesskyScanOption::update_option( $option_name, $wp_version );

            return true;
        }

        return false;
    }

    /**
     * Gather all the checksums (aka. file hashes) of this site, send them, and
     * analyze them using the Dessky Monitoring service, this will generate the
     * audit logs for this site and be part of the integrity checks.
     *
     * @param  boolean $force_scan Whether the filesystem scan was forced by an administrator user or not.
     * @return boolean             TRUE if the filesystem scan was successful, FALSE otherwise.
     */
    public static function filesystem_scan( $force_scan = false ){
        $minimum_runtime = DESSKYSCAN_MINIMUM_RUNTIME;

        if (
            class_exists( 'DesskyScanFileInfo' )
        ) {
            self::report_site_version();

            $file_info = new DesskyScanFileInfo();
            $file_info->scan_interface = DesskyScanOption::get_option( ':scan_interface' );
            $signatures = $file_info->get_directory_tree_md5( ABSPATH );

            if ( $signatures ) {
            } else {
                DesskyScanInterface::error( 'The file hashes could not be retrieved, the filesystem scan failed.' );
            }
        }

        return false;
    }

    /**
     * Generates an audit event log (to be sent later).
     *
     * @param  integer $severity Importance of the event that will be reported, values from one to five.
     * @param  string  $location In which part of the system was the event triggered.
     * @param  string  $message  The explanation of the event.
     * @param  boolean $internal Whether the event will be publicly visible or not.
     * @return boolean           TRUE if the event was logged in the monitoring service, FALSE otherwise.
     */
    private static function report_event( $severity = 0, $location = '', $message = '', $internal = false ){
        $user = wp_get_current_user();
        $username = false;
        $current_time = date( 'Y-m-d H:i:s' );
        $remote_ip = self::get_remote_addr();

        // Identify current user in session.
        if (
            $user instanceof WP_User
            && isset($user->user_login)
            && ! empty($user->user_login)
        ) {
            if ( $user->user_login != $user->display_name ) {
                $username = sprintf( "\x20%s (%s),", $user->display_name, $user->user_login );
            } else {
                $username = sprintf( "\x20%s,", $user->user_login );
            }
        }

        // Fixing severity value.
        $severity = (int) $severity;

        // Convert the severity number into a readable string.
        switch ( $severity ) {
            case 0:  $severity_name = 'Debug';    break;
            case 1:  $severity_name = 'Notice';   break;
            case 2:  $severity_name = 'Info';     break;
            case 3:  $severity_name = 'Warning';  break;
            case 4:  $severity_name = 'Error';    break;
            case 5:  $severity_name = 'Critical'; break;
            default: $severity_name = 'Info';     break;
        }

        // Mark the event as internal if necessary.
        if ( $internal === true ) {
            $severity_name = '@' . $severity_name;
        }

        // Clear event message.
        $message = strip_tags( $message );
        $message = str_replace( "\r", '', $message );
        $message = str_replace( "\n", '', $message );
        $message = str_replace( "\t", '', $message );

        $event_message = sprintf(
            '%s:%s %s; %s',
            $severity_name,
            $username,
            $remote_ip,
            $message
        );
    }

    /**
     * Reports a debug event on the website.
     *
     * @param  string  $message  Text witht the explanation of the event or action performed.
     * @param  boolean $internal Whether the event will be publicly visible or not.
     * @return boolean           Either true or false depending on the success of the operation.
     */
    public static function report_debug_event( $message = '', $internal = false ){
        return self::report_event( 0, 'core', $message, $internal );
    }

    /**
     * Reports a notice event on the website.
     *
     * @param  string  $message  Text witht the explanation of the event or action performed.
     * @param  boolean $internal Whether the event will be publicly visible or not.
     * @return boolean           Either true or false depending on the success of the operation.
     */
    public static function report_notice_event( $message = '', $internal = false ){
        return self::report_event( 1, 'core', $message, $internal );
    }

    /**
     * Reports a info event on the website.
     *
     * @param  string  $message  Text witht the explanation of the event or action performed.
     * @param  boolean $internal Whether the event will be publicly visible or not.
     * @return boolean           Either true or false depending on the success of the operation.
     */
    public static function report_info_event( $message = '', $internal = false ){
        return self::report_event( 2, 'core', $message, $internal );
    }

    /**
     * Reports a error event on the website.
     *
     * @param  string  $message  Text witht the explanation of the event or action performed.
     * @param  boolean $internal Whether the event will be publicly visible or not.
     * @return boolean           Either true or false depending on the success of the operation.
     */
    public static function report_error_event( $message = '', $internal = false ){
        return self::report_event( 4, 'core', $message, $internal );
    }

    /**
     * Reports an esception on the code.
     *
     * @param  Exception $exception A valid exception object of any type.
     * @return boolean              Whether the report was filled correctly or not.
     */
    public static function report_exception( $exception = false ){
        if ( $exception ) {
            $e_trace = $exception->getTrace();
            $multiple_entries = array();

            foreach ( $e_trace as $e_child ) {
                $e_file = array_key_exists( 'file', $e_child )
                    ? basename( $e_child['file'] )
                    : '[internal function]';
                $e_line = array_key_exists( 'line', $e_child )
                    ? basename( $e_child['line'] )
                    : '0';
                $e_function = array_key_exists( 'class', $e_child )
                    ? $e_child['class'] . $e_child['type'] . $e_child['function']
                    : $e_child['function'];
                $multiple_entries[] = sprintf(
                    '%s(%s): %s',
                    $e_file,
                    $e_line,
                    $e_function
                );
            }

            $report_message = sprintf(
                '%s: (multiple entries): %s',
                $exception->getMessage(),
                @implode( ',', $multiple_entries )
            );

            return self::report_debug_event( $report_message );
        }

        return false;
    }

}

/**
 * Read, parse and handle everything related with the templates.
 *
 * A web template system uses a template processor to combine web templates to
 * form finished web pages, possibly using some data source to customize the
 * pages or present a large amount of content on similar-looking pages. It is a
 * web publishing tool present in content management systems, web application
 * frameworks, and HTML editors.
 *
 * Web templates can be used like the template of a form letter to either
 * generate a large number of "static" (unchanging) web pages in advance, or to
 * produce "dynamic" web pages on demand.
 */
class DesskyScanTemplate extends DesskyScanRequest {

    /**
     * Replace all pseudo-variables from a string of characters.
     *
     * @param  string $content The content of a template file which contains pseudo-variables.
     * @param  array  $params  List of pseudo-variables that will be replaced in the template.
     * @return string          The content of the template with the pseudo-variables replated.
     */
    private static function replace_pseudovars( $content = '', $params = array() ){
        if ( is_array( $params ) ) {
            foreach ( $params as $tpl_key => $tpl_value ) {
                $tpl_key = '%%DESSKY.' . $tpl_key . '%%';
                $content = str_replace( $tpl_key, $tpl_value, $content );
            }

            return $content;
        }

        return false;
    }

    /**
     * Gather and generate the information required globally by all the template files.
     *
     * @param  array $params A hash containing the pseudo-variable name as the key and the value that will replace it.
     * @return array         A complementary list of pseudo-variables for the template files.
     */
    private static function shared_params( $params = array() ){
        $params = is_array( $params ) ? $params : array();

        // Base parameters, required to render all the pages.
        $params = self::links_and_navbar( $params );

        // Global parameters, used through out all the pages.
        $params['PageTitle'] = isset($params['PageTitle']) ? ' > '.$params['PageTitle'] : '';
        $params['PageNonce'] = wp_create_nonce( 'desskyscan_page_nonce' );
        $params['PageStyleClass'] = isset($params['PageStyleClass']) ? $params['PageStyleClass'] : 'base';
        $params['CleanDomain'] = self::get_domain();
        $params['LayoutType'] = 'twocolumns';
        $params['AdsVisibility'] = 'visible';
        $params['ReviewNavbarButton'] = 'hidden';

        return $params;
    }

    /**
     * Return a string indicating the visibility of a HTML component.
     *
     * @param  boolean $visible Whether the condition executed returned a positive value or not.
     * @return string           A string indicating the visibility of a HTML component.
     */
    public static function visibility( $visible = false ){
        return ( $visible === true ? 'visible' : 'hidden' );
    }

    /**
     * Generate an URL pointing to the page indicated in the function and that must
     * be loaded through the administrator panel.
     *
     * @param  string $page Short name of the page that will be generated.
     * @return string       Full string containing the link of the page.
     */
    public static function get_url( $page = '' ){
        $url_path = admin_url( 'admin.php?page=desskyscan' );

        if ( ! empty($page) ) {
            $url_path .= '_' . strtolower( $page );
        }

        return $url_path;
    }

    /**
     * Generate an URL pointing to the page indicated in the function and that must
     * be loaded through the Ajax handler of the administrator panel.
     *
     * @param  string $page Short name of the page that will be generated.
     * @return string       Full string containing the link of the page.
     */
    public static function get_ajax_url( $page = '' ){
        $url_path = admin_url( 'admin-ajax.php?page=desskyscan' );

        if ( ! empty($page) ) {
            $url_path .= '_' . strtolower( $page );
        }

        return $url_path;
    }

    /**
     * Complement the list of pseudo-variables that will be used in the base
     * template files, this will also generate the navigation bar and detect which
     * items in it are selected by the current page.
     *
     * @param  array  $params A hash containing the pseudo-variable name as the key and the value that will replace it.
     * @return array          A complementary list of pseudo-variables for the template files.
     */
    private static function links_and_navbar( $params = array() ){
        global $desskyscan_pages;

        $params = is_array( $params ) ? $params : array();
        $sub_pages = is_array( $desskyscan_pages ) ? $desskyscan_pages : array();

        $params['Navbar'] = '';
        $params['CurrentPageFunc'] = '';

        if ( $_page = self::get( 'page', '_page' ) ) {
            $params['CurrentPageFunc'] = $_page;
        }

        foreach ( $sub_pages as $sub_page_func => $sub_page_title ) {
            if (
                $sub_page_func == 'desskyscan_scanner'
                && self::is_sitecheck_disabled()
            ) {
                continue;
            }

            $func_parts = explode( '_', $sub_page_func, 2 );

            if ( isset($func_parts[1]) ) {
                $unique_name = $func_parts[1];
                $pseudo_var = 'URL.' . ucwords( $unique_name );
            } else {
                $unique_name = '';
                $pseudo_var = 'URL.Home';
            }

            $params[ $pseudo_var ] = self::get_url( $unique_name );

            // Copy URL variable and create an Ajax handler.
            $pseudo_var_ajax = 'Ajax' . $pseudo_var;
            $params[ $pseudo_var_ajax ] = self::get_ajax_url( $unique_name );

            $navbar_item_css_class = 'nav-tab';

            if ( $params['CurrentPageFunc'] == $sub_page_func ) {
                $navbar_item_css_class .= chr( 32 ) . 'nav-tab-active';
            }

            $params['Navbar'] .= sprintf(
                '<a class="%s" href="%s">%s</a>' . "\n",
                $navbar_item_css_class,
                $params[ $pseudo_var ],
                $sub_page_title
            );
        }

        return $params;
    }

    /**
     * Generate a HTML code using a template and replacing all the pseudo-variables
     * by the dynamic variables provided by the developer through one of the parameters
     * of the function.
     *
     * @param  string $html   The HTML content of a template file with its pseudo-variables parsed.
     * @param  array  $params A hash containing the pseudo-variable name as the key and the value that will replace it.
     * @return string         The formatted HTML content of the base template.
     */
    public static function get_base_template( $html = '', $params = array() ){
        $params = is_array( $params ) ? $params : array();

        $params = self::shared_params( $params );
        $params['PageContent'] = $html;

        return self::get_template( 'base', $params );
    }

    /**
     * Generate a HTML code using a template and replacing all the pseudo-variables
     * by the dynamic variables provided by the developer through one of the parameters
     * of the function.
     *
     * @param  string  $template Filename of the template that will be used to generate the page.
     * @param  array   $params   A hash containing the pseudo-variable name as the key and the value that will replace it.
     * @param  boolean $type     Either page, section or snippet indicating the type of template that will be retrieved.
     * @return string            The formatted HTML page after replace all the pseudo-variables.
     */
    public static function get_template( $template = '', $params = array(), $type = 'page' ){
        switch ( $type ) {
            case 'page': /* no_break */
            case 'section':
                $template_path_pattern = '%s/%s/assets/tpl/%s.html.tpl';
                break;
            case 'snippet':
                $template_path_pattern = '%s/%s/assets/tpl/%s.snippet.tpl';
                break;
        }

        $template_content = '';
        $template_path = sprintf( $template_path_pattern, WP_PLUGIN_DIR, DESSKYSCAN_PLUGIN_FOLDER, $template );
        $params = is_array( $params ) ? $params : array();

        if ( file_exists( $template_path ) && is_readable( $template_path ) ) {
            $template_content = @file_get_contents( $template_path );

            $params['DesskyURL'] = DESSKYSCAN_URL;

            // Detect the current page URL.
            if ( $_page = self::get( 'page', '_page' ) ) {
                $params['CurrentURL'] = admin_url( 'admin.php?page=' . $_page );
            } else {
                $params['CurrentURL'] = admin_url();
            }

            // Replace the global pseudo-variables in the section/snippets templates.
            if (
                $template == 'base'
                && isset($params['PageContent'])
                && preg_match( '/%%DESSKY\.(.+)%%/', $params['PageContent'] )
            ) {
                $params['PageContent'] = self::replace_pseudovars( $params['PageContent'], $params );
            }

            $template_content = self::replace_pseudovars( $template_content, $params );
        }

        if ( $template == 'base' || $type != 'page' ) {
            return $template_content;
        }

        return self::get_base_template( $template_content, $params );
    }

    /**
     * Generate a HTML code using a template and replacing all the pseudo-variables
     * by the dynamic variables provided by the developer through one of the parameters
     * of the function.
     *
     * @param  string $template Filename of the template that will be used to generate the page.
     * @param  array  $params   A hash containing the pseudo-variable name as the key and the value that will replace it.
     * @return string           The formatted HTML page after replace all the pseudo-variables.
     */
    public static function get_section( $template = '', $params = array() ){
        $params = self::shared_params( $params );

        return self::get_template( $template, $params, 'section' );
    }

    /**
     * Generate a HTML code using a template and replacing all the pseudo-variables
     * by the dynamic variables provided by the developer through one of the parameters
     * of the function.
     *
     * @param  string $template Filename of the template that will be used to generate the page.
     * @param  array  $params   A hash containing the pseudo-variable name as the key and the value that will replace it.
     * @return string           The formatted HTML page after replace all the pseudo-variables.
     */
    public static function get_modal( $template = '', $params = array() ){
        $required = array(
            'Title' => 'Lorem ipsum dolor sit amet',
            'Visibility' => 'visible',
            'Identifier' => 'foobar',
            'CssClass' => '',
            'Content' => '<p>Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do
                eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim
                veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo
                consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse
                cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non
                proident, sunt in culpa qui officia deserunt mollit anim id est laborum.</p>',
        );

        if ( ! empty($template) && $template != 'none' ) {
            $params['Content'] = self::get_section( $template );
        }

        foreach ( $required as $param_name => $param_value ) {
            if ( ! isset($params[ $param_name ]) ) {
                $params[ $param_name ] = $param_value;
            }
        }

        $params['Visibility'] = 'desskyscan-' . $params['Visibility'];
        $params['Identifier'] = 'desskyscan-' . $template . '-modal';
        $params = self::shared_params( $params );

        return self::get_template( 'modalwindow', $params, 'section' );
    }

    /**
     * Generate a HTML code using a template and replacing all the pseudo-variables
     * by the dynamic variables provided by the developer through one of the parameters
     * of the function.
     *
     * @param  string $template Filename of the template that will be used to generate the page.
     * @param  array  $params   A hash containing the pseudo-variable name as the key and the value that will replace it.
     * @return string           The formatted HTML page after replace all the pseudo-variables.
     */
    public static function get_snippet( $template = '', $params = array() ){
        return self::get_template( $template, $params, 'snippet' );
    }

    /**
     * Generate the HTML code necessary to render a list of options in a form.
     *
     * @param  array  $allowed_values List with keys and values allowed for the options.
     * @param  string $selected_val   Value of the option that will be selected by default.
     * @return string                 Option list for a select form field.
     */
    public static function get_select_options( $allowed_values = array(), $selected_val = '' ){
        $options = '';

        foreach ( $allowed_values as $option_name => $option_label ) {
            $selected_str = '';

            if ( $option_name == $selected_val ) {
                $selected_str = 'selected="selected"';
            }

            $options .= sprintf(
                "<option value=\"%s\" %s>%s</option>\n",
                DesskyScan::escape( $option_name ),
                DesskyScan::escape( $selected_str ),
                DesskyScan::escape( $option_label )
            );
        }

        return $options;
    }

    /**
     * Detect which number in a pagination was clicked.
     *
     * @return integer Page number of the link clicked in a pagination.
     */
    public static function get_page_number(){
        $paged = self::get( 'paged', '[0-9]{1,5}' );

        return ( $paged ? intval( $paged ) : 1 );
    }

    /**
     * Check whether the SiteCheck scanner and the malware scan page are disabled.
     *
     * @return boolean TRUE if the SiteCheck scanner and malware scan page are disabled.
     */
    public static function is_sitecheck_disabled(){
        return (bool) ( DesskyScanOption::get_option( ':sitecheck_scanner' ) === 'disabled' );
    }

}

/**
 * Class to process files and folders.
 *
 * Here are implemented the functions needed to open, scan, read, create files
 * and folders using the built-in PHP class SplFileInfo. The SplFileInfo class
 * offers a high-level object oriented interface to information for an individual
 * file.
 */
class DesskyScanFileInfo extends DesskyScan {

    /**
     * Define the interface that will be used to execute the file system scans, the
     * available options are SPL, OpenDir, and Glob (all in lowercase). This can be
     * configured from the settings page.
     *
     * @var string
     */
    public $scan_interface = 'spl';

    /**
     * Whether the list of files that can be ignored from the filesystem scan will
     * be used to return the directory tree, this should be disabled when scanning a
     * directory without the need to filter the items in the list.
     *
     * @var boolean
     */
    public $ignore_files = true;

    /**
     * Whether the list of folders that can be ignored from the filesystem scan will
     * be used to return the directory tree, this should be disabled when scanning a
     * path without the need to filter the items in the list.
     *
     * @var boolean
     */
    public $ignore_directories = true;

    /**
     * A list of ignored directory paths, these folders will be skipped during the
     * execution of the file system scans, and any sub-directory or files inside
     * these paths will be ignored too.
     *
     * @see DesskyScanFSScanner.get_ignored_directories()
     * @var array
     */
    private $ignored_directories = array();

    /**
     * Whether the filesystem scanner should run recursively or not.
     *
     * @var boolean
     */
    public $run_recursively = true;

    /**
     * Whether the directory paths must be skipped or not.
     *
     * This is useful to retrieve the full list of resources inside a parent
     * directory, one case where this option can be set as True is when a folder is
     * required to be deleted recursively, considering that by default the folders
     * are ignored and that a folder may be empty some times there could be issues
     * because the deletion will not reach these resources.
     *
     * @var boolean
     */
    public $skip_directories = true;

    /**
     * Class constructor.
     */
    public function __construct(){
    }

    /**
     * Retrieve a long text string with signatures of all the files contained
     * in the main and subdirectories of the folder specified, also the filesize
     * and md5sum of that file. Some folders and files will be ignored depending
     * on some rules defined by the developer.
     *
     * @param  string  $directory Parent directory where the filesystem scan will start.
     * @param  boolean $as_array  Whether the result of the operation will be returned as an array or string.
     * @return array              List of files in the main and subdirectories of the folder specified.
     */
    public function get_directory_tree_md5( $directory = '', $as_array = false ){
        $project_signatures = '';
        $abs_path = rtrim( ABSPATH, DIRECTORY_SEPARATOR );
        $files = $this->get_directory_tree( $directory );

        if ( $as_array ) {
            $project_signatures = array();
        }

        if ( $files ) {
            sort( $files );

            foreach ( $files as $filepath ) {
                $file_checksum = @md5_file( $filepath );
                $filesize = @filesize( $filepath );

                if ( $as_array ) {
                    $basename = str_replace( $abs_path . DIRECTORY_SEPARATOR, '', $filepath );
                    $project_signatures[ $basename ] = array(
                        'filepath' => $filepath,
                        'checksum' => $file_checksum,
                        'filesize' => $filesize,
                        'created_at' => @filectime( $filepath ),
                        'modified_at' => @filemtime( $filepath ),
                    );
                } else {
                    $filepath = str_replace( $abs_path, $abs_path . DIRECTORY_SEPARATOR, $filepath );
                    $project_signatures .= sprintf(
                        "%s%s%s%s\n",
                        $file_checksum,
                        $filesize,
                        chr( 32 ),
                        $filepath
                    );
                }
            }
        }

        return $project_signatures;
    }

    /**
     * Retrieve a list with all the files contained in the main and subdirectories
     * of the folder specified. Some folders and files will be ignored depending
     * on some rules defined by the developer.
     *
     * @param  string $directory Parent directory where the filesystem scan will start.
     * @return array             List of files in the main and subdirectories of the folder specified.
     */
    public function get_directory_tree( $directory = '' ){
        if ( file_exists( $directory ) && is_dir( $directory ) ) {
            $tree = array();

            // Check whether the ignore scanning feature is enabled or not.
            if ( DesskyScanFSScanner::will_ignore_scanning() ) {
                $this->ignored_directories = DesskyScanFSScanner::get_ignored_directories();
            }

            switch ( $this->scan_interface ) {
                case 'spl':
                    if ( $this->is_spl_available() ) {
                        $tree = $this->get_directory_tree_with_spl( $directory );
                    } else {
                        $this->scan_interface = 'opendir';
                        DesskyScanOption::update_option( ':scan_interface', $this->scan_interface );
                        $tree = $this->get_directory_tree( $directory );
                    }
                    break;

                case 'glob':
                    $tree = $this->get_directory_tree_with_glob( $directory );
                    break;

                case 'opendir':
                    $tree = $this->get_directory_tree_with_opendir( $directory );
                    break;

                default:
                    $this->scan_interface = 'spl';
                    $tree = $this->get_directory_tree( $directory );
                    break;
            }

            return $tree;
        }

        return false;
    }

    /**
     * Check whether the built-in class SplFileObject is available in the system
     * or not, it is required to have PHP >= 5.1.0. The SplFileObject class offers
     * an object oriented interface for a file.
     *
     * @link http://www.php.net/manual/en/class.splfileobject.php
     *
     * @return boolean Whether the PHP class "SplFileObject" is available or not.
     */
    public static function is_spl_available(){
        return (bool) class_exists( 'SplFileObject' );
    }

    /**
     * Retrieve a list with all the files contained in the main and subdirectories
     * of the folder specified. Some folders and files will be ignored depending
     * on some rules defined by the developer.
     *
     * @link http://www.php.net/manual/en/class.recursivedirectoryiterator.php
     * @see  RecursiveDirectoryIterator extends FilesystemIterator
     * @see  FilesystemIterator         extends DirectoryIterator
     * @see  DirectoryIterator          extends SplFileInfo
     * @see  SplFileInfo
     *
     * @param  string $directory Parent directory where the filesystem scan will start.
     * @return array             List of files in the main and subdirectories of the folder specified.
     */
    private function get_directory_tree_with_spl( $directory = '' ){
        $files = array();
        $filepath = @realpath( $directory );
        $objects = array();

        // Exception for directory name must not be empty.
        if ( $filepath === false ) {
            return $files;
        }

        if ( ! class_exists( 'FilesystemIterator' ) ) {
            $this->scan_interface = 'opendir';
            DesskyScanOption::update_option( ':scan_interface', $this->scan_interface );
            $alternative_tree = $this->get_directory_tree( $directory );

            return $alternative_tree;
        }

        try {
            if ( $this->run_recursively ) {
                $flags = FilesystemIterator::KEY_AS_PATHNAME
                    | FilesystemIterator::CURRENT_AS_FILEINFO
                    | FilesystemIterator::SKIP_DOTS
                    | FilesystemIterator::UNIX_PATHS;
                $objects = new RecursiveIteratorIterator(
                    new RecursiveDirectoryIterator( $filepath, $flags ),
                    RecursiveIteratorIterator::SELF_FIRST,
                    RecursiveIteratorIterator::CATCH_GET_CHILD
                );
            } else {
                $objects = new DirectoryIterator( $filepath );
            }
        } catch ( RuntimeException $exception ) {
            DesskyScanEvent::report_exception( $exception );
        }

        foreach ( $objects as $filepath => $fileinfo ) {
            $filename = $fileinfo->getFilename();

            if (
                $this->ignore_folderpath( null, $filename )
                || (
                    $this->skip_directories === true
                    && $fileinfo->isDir()
                )
            ) {
                continue;
            }

            if ( $this->run_recursively ) {
                $directory = dirname( $filepath );
            } else {
                $directory = $fileinfo->getPath();
                $filepath = $directory . '/' . $filename;
            }

            if (
                $this->ignore_folderpath( $directory, $filename )
                || $this->ignore_filepath( $filename )
            ) {
                continue;
            }

            $files[] = $filepath;
        }

        return $files;
    }

    /**
     * Retrieve a list with all the files contained in the main and subdirectories
     * of the folder specified. Some folders and files will be ignored depending
     * on some rules defined by the developer.
     *
     * @param  string $directory Parent directory where the filesystem scan will start.
     * @return array             List of files in the main and subdirectories of the folder specified.
     */
    private function get_directory_tree_with_glob( $directory = '' ){
        $files = array();
        $directory_pattern = sprintf( '%s/*', rtrim( $directory, '/' ) );
        $files_found = @glob( $directory_pattern );

        if ( is_array( $files_found ) ) {
            foreach ( $files_found as $filepath ) {
                $filepath = @realpath( $filepath );
                $directory = dirname( $filepath );
                $filepath_parts = explode( '/', $filepath );
                $filename = array_pop( $filepath_parts );

                if ( is_dir( $filepath ) ) {
                    if ( $this->ignore_folderpath( $directory, $filename ) ) {
                        continue;
                    }

                    if ( $this->run_recursively ) {
                        $sub_files = $this->get_directory_tree_with_glob( $filepath );

                        if ( $sub_files ) {
                            $files = array_merge( $files, $sub_files );
                        }
                    }
                } elseif ( $this->ignore_filepath( $filename ) ) {
                    continue;
                } else {
                    $files[] = $filepath;
                }
            }
        }

        return $files;
    }

    /**
     * Retrieve a list with all the files contained in the main and subdirectories
     * of the folder specified. Some folders and files will be ignored depending
     * on some rules defined by the developer.
     *
     * @param  string $directory Parent directory where the filesystem scan will start.
     * @return array             List of files in the main and subdirectories of the folder specified.
     */
    private function get_directory_tree_with_opendir( $directory = '' ){
        $files = array();
        $dh = @opendir( $directory );

        if ( ! $dh ) {
            return false;
        }

        while ( ($filename = readdir( $dh )) !== false ) {
            $filepath = @realpath( $directory . '/' . $filename );

            if ( $filepath === false ) {
                continue;
            } elseif ( is_dir( $filepath ) ) {
                if ( $this->ignore_folderpath( $directory, $filename ) ) {
                    continue;
                }

                if ( $this->run_recursively ) {
                    $sub_files = $this->get_directory_tree_with_opendir( $filepath );

                    if ( $sub_files ) {
                        $files = array_merge( $files, $sub_files );
                    }
                }
            } else {
                if ( $this->ignore_filepath( $filename ) ) {
                    continue;
                }
                $files[] = $filepath;
            }
        }

        closedir( $dh );
        return $files;
    }

    /**
     * Skip some specific directories and file paths from the filesystem scan.
     *
     * @param  string  $directory Directory where the scanner is located at the moment.
     * @param  string  $filename  Name of the folder or file being scanned at the moment.
     * @return boolean            Either TRUE or FALSE representing that the scan should ignore this folder or not.
     */
    private function ignore_folderpath( $directory = '', $filename = '' ){
        // Ignoring current and parent folders.
        if ( $filename == '.' || $filename == '..' ) {
            return true;
        }

        if ( $this->ignore_directories ) {
            // Ignore directories based on a common regular expression.
            $filepath = @realpath( $directory . '/' . $filename );
            $pattern = '/\/wp-content\/(uploads|cache|backup|w3tc)/';

            if ( preg_match( $pattern, $filepath ) ) {
                return true;
            }

            // Ignore directories specified by the administrator.
            if ( ! empty($this->ignored_directories) ) {
                foreach ( $this->ignored_directories['directories'] as $ignored_dir ) {
                    if (
                        strpos( $directory, $ignored_dir ) !== false
                        || strpos( $filepath, $ignored_dir ) !== false
                    ) {
                        return true;
                    }
                }
            }
        }

        return false;
    }

    /**
     * Skip some specific files from the filesystem scan.
     *
     * @param  string  $filename Name of the folder or file being scanned at the moment.
     * @return boolean           Either TRUE or FALSE representing that the scan should ignore this filename or not.
     */
    private function ignore_filepath( $filename = '' ){
        if ( ! $this->ignore_files ) {
            return false;
        }

        // Ignoring backup files from our clean ups.
        if ( strpos( $filename, '_desskybackup.' ) !== false ) {
            return true;
        }

        // Ignore files specified by the administrator.
        if ( ! empty($this->ignored_directories) ) {
            foreach ( $this->ignored_directories['directories'] as $ignored_dir ) {
                if ( strpos( $ignored_dir, $filename ) !== false ) {
                    return true;
                }
            }
        }

        // Any file maching one of these rules WILL NOT be ignored.
        if (
            ( strpos( $filename, '.php' ) !== false) ||
            ( strpos( $filename, '.htm' ) !== false) ||
            ( strpos( $filename, '.js' ) !== false) ||
            ( strcmp( $filename, '.htaccess' ) == 0     ) ||
            ( strcmp( $filename, 'php.ini' ) == 0     )
        ) {
            return false;
        }

        return true;
    }

    /**
     * Retrieve a list of unique directory paths.
     *
     * @param  array $dir_tree A list of files under a directory.
     * @return array           A list of unique directory paths.
     */
    public function get_diretories_only( $dir_tree = array() ){
        $dirs = array();

        if ( is_string( $dir_tree ) ) {
            $dir_tree = $this->get_directory_tree( $dir_tree );
        }

        if ( is_array( $dir_tree ) && ! empty($dir_tree) ) {
            foreach ( $dir_tree as $filepath ) {
                $dir_path = dirname( $filepath );

                if (
                    ! in_array( $dir_path, $dirs )
                    && ! in_array( $dir_path, $this->ignored_directories['directories'] )
                ) {
                    $dirs[] = $dir_path;
                }
            }
        }

        return $dirs;
    }

    /**
     * Return the lines of a file as an array, it will automatically remove the new
     * line characters from the end of each line, and skip empty lines from the
     * list.
     *
     * @param  string $filepath Path to the file.
     * @return array            An array where each element is a line in the file.
     */
    public static function file_lines( $filepath = '' ){
        return @file( $filepath, FILE_IGNORE_NEW_LINES | FILE_SKIP_EMPTY_LINES );
    }

    /**
     * Function to emulate the UNIX tail function by displaying the last X number of
     * lines in a file. Useful for large files, such as logs, when you want to
     * process lines in PHP or write lines to a database.
     *
     * @param  string  $file_path Path to the file.
     * @param  integer $lines     Number of lines to retrieve from the end of the file.
     * @param  boolean $adaptive  Whether the buffer will adapt to a specific number of bytes or not.
     * @return string             Text contained at the end of the file.
     */
    public static function tail_file( $file_path = '', $lines = 1, $adaptive = true ){
        $file = @fopen( $file_path, 'rb' );
        $limit = $lines;

        if ( $file ) {
            fseek( $file, -1, SEEK_END );

            if ( $adaptive && $lines < 2  ) {
                $buffer = 64;
            } elseif ( $adaptive && $lines < 10 ) {
                $buffer = 512;
            } else {
                $buffer = 4096;
            }

            if ( fread( $file, 1 ) != "\n" ) {
                $lines -= 1;
            }

            $output = '';
            $chunk = '';

            while ( ftell( $file ) > 0 && $lines >= 0 ) {
                $seek = min( ftell( $file ), $buffer );
                fseek( $file, -$seek, SEEK_CUR );
                $chunk = fread( $file, $seek );
                $output = $chunk . $output;
                fseek( $file, -mb_strlen( $chunk, '8bit' ), SEEK_CUR );
                $lines -= substr_count( $chunk, "\n" );
            }

            fclose( $file );

            $lines_arr = explode( "\n", $output );
            $lines_count = count( $lines_arr );
            $result = array_slice( $lines_arr, ($lines_count - $limit) );

            return $result;
        }

        return false;
    }

    /**
     * Tells whether the filename is a directory, symbolic link, or file.
     *
     * @param  string $path Path to the file.
     * @return string       Type of resource: dir, link, file.
     */
    public static function get_resource_type( $path = '' ){
        if ( is_dir( $path ) ) {
            return 'dir';
        } elseif ( is_link( $path ) ) {
            return 'link';
        } elseif ( is_file( $path ) ) {
            return 'file';
        } else {
            return 'unknown';
        }
    }

}

/**
 * Plugin initializer.
 *
 * Define all the required variables, script, styles, and basic functions needed
 * when the site is loaded, not even the administrator panel but also the front
 * page, some bug-fixes will/are applied here for sites behind a proxy, and
 * sites with old versions of the premium plugin (that was deprecated at
 * July/2014).
 */
class DesskyScanInterface {

    /**
     * Define which javascript and css files will be loaded in the header of the
     * plugin pages, only when the administrator panel is accessed.
     *
     * @return void
     */
    public static function enqueue_scripts(){
        $asset_version = '';

        if ( strlen( DESSKYSCAN_PLUGIN_CHECKSUM ) >= 7 ) {
            $asset_version = substr( DESSKYSCAN_PLUGIN_CHECKSUM, 0, 7 );
        }

        wp_register_style( 'desskyscan', DESSKYSCAN_URL . '/assets/css/dessky-security.css', array(), $asset_version );
        wp_register_script( 'desskyscan', DESSKYSCAN_URL . '/assets/js/dessky-security.js', array(), $asset_version );
        wp_enqueue_style( 'desskyscan' );
        wp_enqueue_script( 'desskyscan' );
        
    }

    /**
     * Generate the menu and submenus for the plugin in the admin interface.
     *
     * @return void
     */
    public static function add_interface_menu(){
        global $desskyscan_pages;

        if (
            function_exists( 'add_menu_page' )
            && $desskyscan_pages
            && is_array( $desskyscan_pages )
            && array_key_exists( 'desskyscan_hardening', $desskyscan_pages )
        ) {
            // Add main menu link.
            add_menu_page(
                'Dessky Security',
                'Dessky Security',
                'manage_options',
                'desskyscan_hardening',
                'desskyscan_hardening_page',
                'dashicons-shield-alt'
            );

            foreach ( $desskyscan_pages as $sub_page_func => $sub_page_title ) {
                if (
                    $sub_page_func == 'desskyscan_scanner'
                    && DesskyScanTemplate::is_sitecheck_disabled()
                ) {
                    continue;
                }

                $page_func = $sub_page_func . '_page';

                add_submenu_page(
                    'desskyscan_hardening',
                    $sub_page_title,
                    $sub_page_title,
                    'manage_options',
                    $sub_page_func,
                    $page_func
                );
            }
        }
    }

    /**
     * Check whether a user has the permissions to see a page from the plugin.
     *
     * @return void
     */
    public static function check_permissions(){
        if (
            ! function_exists( 'current_user_can' )
            || ! current_user_can( 'manage_options' )
        ) {
            $page = DesskyScanRequest::get( 'page', '_page' );
            wp_die( __( 'Access denied by <b>Dessky</b> to see <code>' . $page . '</code>' ) );
        }
    }

    /**
     * Verify the nonce of the previous page after a form submission. If the
     * validation fails the execution of the script will be stopped and a dead page
     * will be printed to the client using the official WordPress method.
     *
     * @return boolean Either TRUE or FALSE if the nonce is valid or not respectively.
     */
    public static function check_nonce(){
        if ( ! empty($_POST) ) {
            $nonce_name = 'desskyscan_page_nonce';
            $nonce_value = DesskyScanRequest::post( $nonce_name, '_nonce' );

            if ( ! $nonce_value || ! wp_verify_nonce( $nonce_value, $nonce_name ) ) {
                wp_die( __( 'WordPress Nonce verification failed, try again going back and checking the form.' ) );

                return false;
            }
        }

        return true;
    }

    /**
     * Prints a HTML alert in the WordPress admin interface.
     *
     * @param  string $type    The type of alert, it can be either Updated or Error.
     * @param  string $message The message that will be printed in the alert.
     * @return void
     */
    private static function admin_notice( $type = 'updated', $message = '' ){
        $display_notice = true;

        /**
         * Do not render notice during user authentication.
         *
         * There are some special cases when the error or warning messages should not be
         * rendered to the end user because it may break the default functionality of
         * the request handler. For instance, rendering an HTML alert like this when the
         * user authentication process is executed may cause a "headers already sent"
         * error.
         */
        if (
            ! empty( $_POST )
            && DesskyScanRequest::post( 'log' )
            && DesskyScanRequest::post( 'pwd' )
            && DesskyScanRequest::post( 'wp-submit' )
        ) {
            $display_notice = false;
        }

        // Display the HTML notice to the current user.
        if ( $display_notice === true ) {
            $alert_id = rand( 100, 999 );
            if ( ! empty($message) ): ?>
                <div id="desskyscan-alert-<?php echo $alert_id; ?>" class="<?php echo $type; ?> desskyscan-alert desskyscan-alert-<?php echo $type; ?>">
                    <a href="javascript:void(0)" class="close" onclick="desskyscan_alert_close('<?php echo $alert_id; ?>')">&times;</a>
                    <p><?php _e( $message ); ?></p>
                </div>
            <?php endif;
        }
    }

    /**
     * Prints a HTML alert of type ERROR in the WordPress admin interface.
     *
     * @param  string $error_msg The message that will be printed in the alert.
     * @return void
     */
    public static function error( $error_msg = '' ){
        self::admin_notice( 'error', '<b>Dessky:</b> ' . $error_msg );
    }

    /**
     * Prints a HTML alert of type INFO in the WordPress admin interface.
     *
     * @param  string $info_msg The message that will be printed in the alert.
     * @return void
     */
    public static function info( $info_msg = '' ){
        self::admin_notice( 'updated', '<b>Dessky:</b> ' . $info_msg );
    }
}

/**
 * Project hardening library.
 *
 * In computing, hardening is usually the process of securing a system by
 * reducing its surface of vulnerability. A system has a larger vulnerability
 * surface the more functions it fulfills; in principle a single-function system
 * is more secure than a multipurpose one. Reducing available vectors of attack
 * typically includes the removal of unnecessary software, unnecessary usernames
 * or logins and the disabling or removal of unnecessary services.
 *
 * There are various methods of hardening Unix and Linux systems. This may
 * involve, among other measures, applying a patch to the kernel such as Exec
 * Shield or PaX; closing open network ports; and setting up intrusion-detection
 * systems, firewalls and intrusion-prevention systems. There are also hardening
 * scripts and tools like Bastille Linux, JASS for Solaris systems and
 * Apache/PHP Hardener that can, for example, deactivate unneeded features in
 * configuration files or perform various other protective measures.
 */
class DesskyScanHardening extends DesskyScan {

    /**
     * Returns a list of access control rules for the Apache web server that can be
     * used to deny and allow certain files to be accessed by certain network nodes.
     * Currently supports Apache 2.2 and 2.4 and denies access to all PHP files with
     * any mixed extension case.
     *
     * @param  string $directory Valid directory path.
     * @return array             List of access control rules.
     */
    private static function get_rules( $directory = '' ){
        $directory = basename( $directory );
        $rules = array(
            '<FilesMatch "\.(?i:php)$">',
            '  <IfModule !mod_authz_core.c>',
            '    Order allow,deny',
            '    Deny from all',
            '  </IfModule>',
            '  <IfModule mod_authz_core.c>',
            '    Require all denied',
            '  </IfModule>',
            '</FilesMatch>',
        );

        if ( $directory == 'wp-includes' ) {
            $rules[] = '<Files wp-tinymce.php>';
            $rules[] = '  Allow from all';
            $rules[] = '</Files>';
            $rules[] = '<Files ms-files.php>';
            $rules[] = '  Allow from all';
            $rules[] = '</Files>';
        }

        return $rules;
    }

    /**
     * Adds some rules to an existing access control file (or creates it if does not
     * exists) to deny access to all files with certain extension in any mixed case.
     * The permissions to modify the file are checked before anything else, this
     * function is self-contained.
     *
     * @param  string  $directory Valid directory path where to place the access rules.
     * @return boolean            True if the rules are successfully added, false otherwise.
     */
    public static function harden_directory( $directory = '' ){
        if (
            file_exists( $directory)
            && is_writable( $directory )
            && is_dir( $directory )
        ) {
            $fhandle = false;
            $target = $directory . '/.htaccess';
            $deny_rules = self::get_rules( $directory );

            if ( file_exists( $target ) ) {
                self::fix_previous_hardening( $directory );
                $fhandle = @fopen( $target, 'a' );
            } else {
                $fhandle = @fopen( $target, 'w' );
            }

            if ( $fhandle ) {
                $rules_str = implode( "\n", $deny_rules );
                $written = fwrite( $fhandle, $rules_str );
                fclose( $fhandle );

                return (bool) ( $written !== false );
            }
        }

        return false;
    }

    /**
     * Deletes some rules from an existing access control file to allow access to
     * all files with certain extension in any mixed case. The file is truncated if
     * after the operation its size is equals to zero.
     *
     * @param  string  $directory Valid directory path where to access rules are.
     * @return boolean            True if the rules are successfully deleted, false otherwise.
     */
    public static function unharden_directory( $directory = '' ){
        if ( self::is_hardened( $directory ) ) {
            $deny_rules = self::get_rules( $directory );
            $fpath = $directory . '/.htaccess';
            $content = @file_get_contents( $fpath );

            if ( $content ) {
                $rules_str = implode( "\n", $deny_rules );
                $content = str_replace( $rules_str, '', $content );
                $written = @file_put_contents( $fpath, $content );

                if ( filesize( $fpath ) === 0 ) {
                    @unlink( $fpath );
                }

                return (bool) ( $written !== false );
            }
        }

        return false;
    }

    /**
     * Remove the hardening applied in previous versions.
     *
     * @param  string  $directory Valid directory path.
     * @return boolean            True if the access control file was fixed.
     */
    private static function fix_previous_hardening( $directory = '' ){
        $fpath = $directory . '/.htaccess';
        $content = @file_get_contents( $fpath );
        $rules = "<Files *.php>\ndeny from all\n</Files>";

        if ( $content ) {
            if ( strpos( $content, $rules ) !== false ) {
                $content = str_replace( $rules, '', $content );
                $written = @file_put_contents( $fpath, $content );

                return (bool) ( $written !== false );
            }
        }

        return true;
    }

    /**
     * Check whether a directory is hardened or not.
     *
     * @param  string  $directory Valid directory path.
     * @return boolean            True if the directory is hardened, false otherwise.
     */
    public static function is_hardened( $directory = '' ){
        if (
            file_exists( $directory )
            && is_dir( $directory )
        ) {
            $fpath = $directory . '/.htaccess';

            if (
                file_exists( $fpath )
                && is_readable( $fpath )
            ) {
                $rules = self::get_rules( $directory );
                $rules_str = implode( "\n", $rules );
                $content = @file_get_contents( $fpath );

                if ( strpos( $content, $rules_str ) !== false ) {
                    return true;
                }
            }
        }

        return false;
    }

}

/**
 * HTTP request handler.
 *
 * Function definitions to retrieve, validate, and clean the parameters during a
 * HTTP request, generally after a form submission or while loading a URL. Use
 * these methods at most instead of accessing an index in the global PHP
 * variables _POST, _GET, _REQUEST since they may come with insecure data.
 */
class DesskyScanRequest extends DesskyScan {

    /**
     * Returns the value stored in a specific index in the global _GET, _POST or
     * _REQUEST variables, you can specify a pattern as the second argument to
     * match allowed values.
     *
     * @param  array  $list    The array where the specified key will be searched.
     * @param  string $key     Name of the index where the requested variable is supposed to be.
     * @param  string $pattern Optional pattern to match allowed values in the requested key.
     * @return string          The value stored in the specified key inside the global _GET variable.
     */
    public static function request( $list = array(), $key = '', $pattern = '' ){
        $key = self::variable_prefix( $key );

        if (
            is_array( $list )
            && is_string( $key )
            && isset($list[ $key ])
        ) {
            // Select the key from the list and escape its content.
            $key_value = $list[ $key ];

            // Define regular expressions for specific value types.
            if ( $pattern === '' ) {
                $pattern = '/.*/';
            } else {
                switch ( $pattern ) {
                    case '_nonce': $pattern = '/^[a-z0-9]{10}$/'; break;
                    case '_page': $pattern = '/^[a-z_]+$/'; break;
                    case '_array': $pattern = '_array'; break;
                    case '_yyyymmdd': $pattern = '/^[0-9]{4}(\-[0-9]{2}) {2}$/'; break;
                    default: $pattern = '/^'.$pattern.'$/'; break;
                }
            }

            // If the request data is an array, then only cast the value.
            if ( $pattern == '_array' && is_array( $key_value ) ) {
                return (array) $key_value;
            }

            // Check the format of the request data with a regex defined above.
            if ( @preg_match( $pattern, $key_value ) ) {
                return self::escape( $key_value );
            }
        }

        return false;
    }

    /**
     * Returns the value stored in a specific index in the global _GET variable,
     * you can specify a pattern as the second argument to match allowed values.
     *
     * @param  string $key     Name of the index where the requested variable is supposed to be.
     * @param  string $pattern Optional pattern to match allowed values in the requested key.
     * @return string          The value stored in the specified key inside the global _GET variable.
     */
    public static function get( $key = '', $pattern = '' ){
        return self::request( $_GET, $key, $pattern );
    }

    /**
     * Returns the value stored in a specific index in the global _POST variable,
     * you can specify a pattern as the second argument to match allowed values.
     *
     * @param  string $key     Name of the index where the requested variable is supposed to be.
     * @param  string $pattern Optional pattern to match allowed values in the requested key.
     * @return string          The value stored in the specified key inside the global _POST variable.
     */
    public static function post( $key = '', $pattern = '' ){
        return self::request( $_POST, $key, $pattern );
    }

}

/**
 * File System Scanner
 *
 * The File System Scanner component performs full and incremental scans over a
 * file system folder, maintaining a snapshot of the filesystem and comparing it
 * with the current content to establish what content has been updated. Updated
 * content is then submitted to the remote server and it is stored for future
 * analysis.
 */
class DesskyScanFSScanner extends DesskyScan {

    /**
     * Check whether the administrator enabled the feature to ignore some
     * directories during the file system scans. This function is overwritten by a
     * GET parameter in the settings page named no_scan which must be equal to the
     * number one.
     *
     * @return boolean Whether the feature to ignore files is enabled or not.
     */
    public static function will_ignore_scanning(){
        return ( DesskyScanOption::get_option( ':ignore_scanning' ) === 'enabled' );
    }

    /**
     * Add a new directory path to the list of ignored paths.
     *
     * @param  string  $directory_path The (full) absolute path of a directory.
     * @return boolean                 TRUE if the directory path was added to the list, FALSE otherwise.
     */
    public static function ignore_directory( $directory_path = '' ){
        $cache = new DesskyScanCache( 'ignorescanning' );

        // Use the checksum of the directory path as the cache key.
        $cache_key = md5( $directory_path );
        $resource_type = DesskyScanFileInfo::get_resource_type( $directory_path );
        $cache_value = array(
            'directory_path' => $directory_path,
            'ignored_at' => self::local_time(),
            'resource_type' => $resource_type,
        );
        $cached = $cache->add( $cache_key, $cache_value );

        return $cached;
    }

    /**
     * Retrieve a list of directories ignored.
     *
     * Retrieve a list of directory paths that will be ignored during the file
     * system scans, any sub-directory and files inside these folders will be
     * skipped automatically and will not be used to detect malware or modifications
     * in the site.
     *
     * The structure of the array returned by the function will always be composed
     * by four (4) indexes which will facilitate the execution of common conditions
     * in the implementation code.
     *
     * <ul>
     * <li>raw: Will contains the raw data retrieved from the built-in cache system.</li>
     * <li>checksums: Will contains the md5 of all the directory paths.</li>
     * <li>directories: Will contains a list of directory paths.</li>
     * <li>ignored_at_list: Will contains a list of timestamps for when the directories were ignored.</li>
     * </ul>
     *
     * @return array List of ignored directory paths.
     */
    public static function get_ignored_directories(){
        $response = array(
            'raw' => array(),
            'checksums' => array(),
            'directories' => array(),
            'ignored_at_list' => array(),
        );

        $cache = new DesskyScanCache( 'ignorescanning' );
        $cache_lifetime = 0; // It is not necessary to expire this cache.
        $ignored_directories = $cache->get_all( $cache_lifetime, 'array' );

        if ( $ignored_directories ) {
            $response['raw'] = $ignored_directories;

            foreach ( $ignored_directories as $checksum => $data ) {
                if (
                    array_key_exists( 'directory_path', $data )
                    && array_key_exists( 'ignored_at', $data )
                ) {
                    $response['checksums'][] = $checksum;
                    $response['directories'][] = $data['directory_path'];
                    $response['ignored_at_list'][] = $data['ignored_at'];
                }
            }
        }

        return $response;
    }

    /**
     * Read and parse the lines inside a PHP error log file.
     *
     * @param  array $error_logs The content of an error log file, or an array with the lines.
     * @return array             List of valid error logs with their attributes separated.
     */
    public static function parse_error_logs( $error_logs = array() ){
        $logs_arr = array();
        $pattern = '/^'
            . '(\[(\S+) ([0-9:]{5,8})( \S+)?\] )?' // Detect date, time, and timezone.
            . '(PHP )?([a-zA-Z ]+):\s'             // Detect PHP error severity.
            . '(.+) in (.+)'                       // Detect error message, and file path.
            . '(:| on line )([0-9]+)'              // Detect line number.
            . '$/';

        if ( is_string( $error_logs ) ) {
            $error_logs = explode( "\n", $error_logs );
        }

        foreach ( (array) $error_logs as $line ) {
            if ( ! is_string( $line ) || empty($line) ) {
                continue;
            }

            if ( preg_match( $pattern, $line, $match ) ) {
                $data_set = array(
                    'date' => '',
                    'time' => '',
                    'timestamp' => 0,
                    'date_time' => '',
                    'time_zone' => '',
                    'error_type' => '',
                    'error_code' => 'unknown',
                    'error_message' => '',
                    'file_path' => '',
                    'line_number' => 0,
                );

                // Basic attributes from the scrapping.
                $data_set['date'] = $match[2];
                $data_set['time'] = $match[3];
                $data_set['time_zone'] = trim( $match[4] );
                $data_set['error_type'] = trim( $match[6] );
                $data_set['error_message'] = trim( $match[7] );
                $data_set['file_path'] = trim( $match[8] );
                $data_set['line_number'] = (int) $match[10];

                // Additional data from the attributes.
                if ( $data_set['date'] ) {
                    $data_set['date_time'] = $data_set['date']
                        . "\x20" . $data_set['time']
                        . "\x20" . $data_set['time_zone'];
                    $data_set['timestamp'] = strtotime( $data_set['date_time'] );
                }

                if ( $data_set['error_type'] ) {
                    $valid_types = array( 'warning', 'notice', 'error' );

                    foreach ( $valid_types as $valid_type ) {
                        if ( stripos( $data_set['error_type'], $valid_type ) !== false ) {
                            $data_set['error_code'] = $valid_type;
                            break;
                        }
                    }
                }

                $logs_arr[] = (object) $data_set;
            }
        }

        return $logs_arr;
    }

}

/**
 * Plugin options handler.
 *
 * Options are pieces of data that WordPress uses to store various preferences
 * and configuration settings. Listed below are the options, along with some of
 * the default values from the current WordPress install. By using the
 * appropriate function, options can be added, changed, removed, and retrieved,
 * from the wp_options table.
 *
 * The Options API is a simple and standardized way of storing data in the
 * database. The API makes it easy to create, access, update, and delete
 * options. All the data is stored in the wp_options table under a given custom
 * name. This page contains the technical documentation needed to use the
 * Options API. A list of default options can be found in the Option Reference.
 *
 * Note that the _site_ functions are essentially the same as their
 * counterparts. The only differences occur for WP Multisite, when the options
 * apply network-wide and the data is stored in the wp_sitemeta table under the
 * given custom name.
 *
 * @see http://codex.wordpress.org/Option_Reference
 * @see http://codex.wordpress.org/Options_API
 */
class DesskyScanOption extends DesskyScanRequest {

    /**
     * Default values for all the plugin's options.
     *
     * @return array Default values for all the plugin's options.
     */
    public static function get_default_option_values(){
        $defaults = array(
            'desskyscan_account' => '',
            'desskyscan_api_key' => false,
            'desskyscan_datastore_path' => '',
            'desskyscan_dns_lookups' => 'enabled',
            'desskyscan_errorlogs_limit' => 30,
            'desskyscan_fs_scanner' => 'enabled',
            'desskyscan_ignore_scanning' => 'disabled',
            'desskyscan_parse_errorlogs' => 'enabled',
            'desskyscan_runtime' => 0,
            'desskyscan_scan_checksums' => 'enabled',
			'desskyscan_scan_errorlogs' => 'disabled',
			'desskyscan_scan_interface' => 'spl',
            'desskyscan_site_version' => '0.0',
            'desskyscan_sitecheck_scanner' => 'enabled',
            'desskyscan_verify_ssl_cert' => 'false',
        );

        return $defaults;
    }

    /**
     * Name of all valid plugin's options.
     *
     * @return array Name of all valid plugin's options.
     */
    public static function get_default_option_names(){
        $options = self::get_default_option_values();
        $names = array_keys( $options );

        return $names;
    }

    /**
     * Retrieve the default values for some specific options.
     *
     * @param  string|array $settings Either an array that will be complemented or a string with the name of the option.
     * @return string|array           The default values for the specified options.
     */
    public static function get_default_options( $settings = '' ){
        $default_options = self::get_default_option_values();

        if ( is_array( $settings ) ) {
            foreach ( $default_options as $option_name => $option_value ) {
                if ( ! isset($settings[ $option_name ]) ) {
                    $settings[ $option_name ] = $option_value;
                }
            }

            return $settings;
        }

        if (
            is_string( $settings )
            && ! empty($settings)
            && array_key_exists( $settings, $default_options )
        ) {
            return $default_options[ $settings ];
        }

        return false;
    }

    /**
     * Alias function for the method Common::DesskyScan_Get_Options()
     *
     * This function search the specified option in the database, not only the options
     * set by the plugin but all the options set for the site. If the value retrieved
     * is FALSE the method tries to search for a default value.
     *
     * To facilitate the development, you can prefix the name of the key in the
     * request (when accessing it) with a single colon, this function will
     * automatically replace that character with the unique identifier of the
     * plugin.
     *
     * @see http://codex.wordpress.org/Function_Reference/get_option
     *
     * @param  string $option_name Optional parameter that you can use to filter the results to one option.
     * @return string              The value (or default value) of the option specified.
     */
    public static function get_option( $option_name = '' ){
        if ( function_exists( 'update_option' ) ) {
            $option_name = self::variable_prefix( $option_name );
            $option_value = get_option( $option_name );

            if ( $option_value === false && preg_match( '/^desskyscan_/', $option_name ) ) {
                $option_value = self::get_default_options( $option_name );
            }

            return $option_value;
        }

        return false;
    }

    /**
     * Update the value of an database' option.
     *
     * Use the function to update a named option/value pair to the options database
     * table. The option name value is escaped with a special database method before
     * the insert SQL statement but not the option value, this value should always
     * be properly sanitized.
     *
     * @see http://codex.wordpress.org/Function_Reference/update_option
     *
     * @param  string  $option_name  Name of the option to update which must not exceed 64 characters.
     * @param  string  $option_value The new value for the option, can be an integer, string, array, or object.
     * @return boolean               True if option value has changed, false if not or if update failed.
     */
    public static function update_option( $option_name = '', $option_value = '' ){
        if ( function_exists( 'update_option' ) ) {
            $option_name = self::variable_prefix( $option_name );

            return update_option( $option_name, $option_value );
        }

        return false;
    }

    /**
     * Remove an option from the database.
     *
     * A safe way of removing a named option/value pair from the options database table.
     *
     * @see http://codex.wordpress.org/Function_Reference/delete_option
     *
     * @param  string  $option_name Name of the option to be deleted.
     * @return boolean              True, if option is successfully deleted. False on failure, or option does not exist.
     */
    public static function delete_option( $option_name = '' ){
        if ( function_exists( 'delete_option' ) ) {
            $option_name = self::variable_prefix( $option_name );

            return delete_option( $option_name );
        }

        return false;
    }

    /**
     * Retrieve all the options stored by Wordpress in the database. The options
     * containing the word "transient" are excluded from the results, this function
     * is compatible with multisite instances.
     *
     * @return array All the options stored by Wordpress in the database, except the transient options.
     */
    public static function get_site_options(){
        global $wpdb;

        $settings = array();
        $results = $wpdb->get_results(
            "SELECT * FROM {$wpdb->options}
            WHERE option_name NOT LIKE '%_transient_%'
            ORDER BY option_id ASC"
        );

        foreach ( $results as $row ) {
            $settings[ $row->option_name ] = $row->option_value;
        }

        return $settings;
    }

    /**
     * Retrieve a list of basic security keys and check whether their values were
     * randomized correctly.
     *
     * @return array Array with three keys: good, missing, bad.
     */
    public static function get_security_keys(){
        $response = array(
            'good' => array(),
            'missing' => array(),
            'bad' => array(),
        );
        $key_names = array(
            'AUTH_KEY',
            'AUTH_SALT',
            'LOGGED_IN_KEY',
            'LOGGED_IN_SALT',
            'NONCE_KEY',
            'NONCE_SALT',
            'SECURE_AUTH_KEY',
            'SECURE_AUTH_SALT',
        );

        foreach ( $key_names as $key_name ) {
            if ( defined( $key_name ) ) {
                $key_value = constant( $key_name );

                if ( stripos( $key_value, 'unique phrase' ) !== false ) {
                    $response['bad'][ $key_name ] = $key_value;
                } else {
                    $response['good'][ $key_name ] = $key_value;
                }
            } else {
                $response['missing'][ $key_name ] = false;
            }
        }

        return $response;
    }

}


/**
 * Dessky one-click hardening page.
 *
 * It loads all the functions defined in /lib/hardening.php and shows the forms
 * that the administrator can use to harden multiple parts of the site.
 *
 * @return void
 */
function desskyscan_hardening_page(){
    DesskyScanInterface::check_permissions();

    if (
        DesskyScanRequest::post( ':run_hardening' )
        && ! DesskyScanInterface::check_nonce()
    ) {
        unset($_POST['desskyscan_run_hardening']);
    }

    $template_variables = array(
        'PageTitle' => 'Hardening',
        'Hardening.Version' => desskyscan_harden_version(),
        'Hardening.RemoveGenerator' => desskyscan_harden_removegenerator(),
        'Hardening.NginxPhpFpm' => '',
        'Hardening.Upload' => '',
        'Hardening.WpContent' => '',
        'Hardening.WpIncludes' => '',
        'Hardening.PhpVersion' => desskyscan_harden_phpversion(),
        'Hardening.SecretKeys' => desskyscan_harden_secretkeys(),
        'Hardening.Readme' => desskyscan_harden_readme(),
        'Hardening.AdminUser' => desskyscan_harden_adminuser(),
        'Hardening.FileEditor' => desskyscan_harden_fileeditor(),
        'Hardening.DBTables' => desskyscan_harden_dbtables(),
    );

    if ( DesskyScan::is_nginx_server() === true ) {
        $template_variables['Hardening.NginxPhpFpm'] = desskyscan_harden_nginx_phpfpm();
    } elseif ( DesskyScan::is_iis_server() === true ) {
        /* TODO: Include IIS (Internet Information Services) hardening options. */
    } else {
        $template_variables['Hardening.Upload'] = desskyscan_harden_upload();
        $template_variables['Hardening.WpContent'] = desskyscan_harden_wpcontent();
        $template_variables['Hardening.WpIncludes'] = desskyscan_harden_wpincludes();
    }

    echo DesskyScanTemplate::get_template( 'hardening', $template_variables );
}

/**
 * Check whether the version number of the WordPress installed is the latest
 * version available officially.
 *
 * @return void
 */
function desskyscan_harden_version(){
    $site_version = DesskyScan::site_version();
    $updates = get_core_updates();
    $cp = ( ! is_array( $updates ) || empty($updates) ? 1 : 0 );

    if ( isset($updates[0]) && $updates[0] instanceof stdClass ) {
        if (
            $updates[0]->response == 'latest'
            || $updates[0]->response == 'development'
        ) {
            $cp = 1;
        }
    }

    if ( strcmp( $site_version, '3.7' ) < 0 ) {
        $cp = 0;
    }

    $initial_msg = 'Why keep your site updated? WordPress is an open-source
        project which means that with every update the details of the changes made
        to the source code are made public, if there were security fixes then
        someone with malicious intent can use this information to attack any site
        that has not been upgraded.';
    $messageok = sprintf( 'Your WordPress installation (%s) is current.', $site_version );
    $messagewarn = sprintf(
        'Your current version (%s) is not current.<br>
        <a href="update-core.php" class="button-primary">Update now!</a>',
        $site_version
    );

    return desskyscan_harden_status( 'Verify WordPress version', $cp, null, $messageok, $messagewarn, $initial_msg );
}

/**
 * Notify the state of the hardening for the removal of the Generator tag in
 * HTML code printed by WordPress to show the current version number of the
 * installation.
 *
 * @return void
 */
function desskyscan_harden_removegenerator(){
    return desskyscan_harden_status(
        'Remove WordPress version',
        1,
        null,
        'WordPress version properly hidden',
        null,
        'It checks if your WordPress version is being hidden from being displayed '
        .'in the generator tag (enabled by default with this plugin).'
    );
}

/**
 * Check the version number of the PHP interpreter set to work with the site,
 * is considered that old versions of the PHP interpreter are insecure.
 *
 * @return void
 */
function desskyscan_harden_phpversion(){
    $phpv = phpversion();
    $cp = ( strncmp( $phpv, '5.', 2 ) < 0 ) ? 0 : 1;

    return desskyscan_harden_status(
        'Verify PHP version',
        $cp,
        null,
        'Using an updated version of PHP (' . $phpv . ')',
        'The version of PHP you are using (' . $phpv . ') is not current, not recommended, and/or not supported',
        'This checks if you have the latest version of PHP installed.',
        null
    );
}

/**
 * Check whether the Wordpress configuration file has the security keys recommended
 * to avoid any unauthorized access to the interface.
 *
 * WordPress Security Keys is a set of random variables that improve encryption of
 * information stored in the user’s cookies. There are a total of four security
 * keys: AUTH_KEY, SECURE_AUTH_KEY, LOGGED_IN_KEY, and NONCE_KEY.
 *
 * @return void
 */
function desskyscan_harden_secretkeys(){
    $wp_config_path = DesskyScan::get_wpconfig_path();
    $current_keys = DesskyScanOption::get_security_keys();

    if ( $wp_config_path ) {
        $cp = 1;
        $message = 'The main configuration file was found at: <code>'.$wp_config_path.'</code><br>';

        if (
            ! empty($current_keys['bad'])
            || ! empty($current_keys['missing'])
        ) {
            $cp = 0;
        }
    } else {
        $cp = 0;
        $message = 'The <code>wp-config.php</code> file was not found.<br>';
    }

    $message .= '<br>It checks whether you have proper random keys/salts created for WordPress. A
        <a href="http://codex.wordpress.org/Editing_wp-config.php#Security_Keys" target="_blank">
        secret key</a> makes your site harder to hack and access harder to crack by adding
        random elements to the password. In simple terms, a secret key is a password with
        elements that make it harder to generate enough options to break through your
        security barriers.';
    $messageok = 'Security keys and salts not set, we recommend to create them for security reasons'
        . '<a href="' . DesskyScanTemplate::get_url( 'posthack' ) . '" class="button button-primary">'
        . 'Harden</a>';

    return desskyscan_harden_status(
        'Security keys',
        $cp,
        null,
        'Security keys and salts properly created',
        $messageok,
        $message,
        null
    );
}

/**
 * Check whether the "readme.html" file is still available in the root of the
 * site or not, which can lead to an attacker to know which version number of
 * Wordpress is being used and search for possible vulnerabilities.
 *
 * @return void
 */
function desskyscan_harden_readme(){
    $upmsg = null;
    $cp = is_readable( ABSPATH.'/readme.html' ) ? 0 : 1;

    // TODO: After hardening create an option to automatically remove this after WP upgrade.
    if ( DesskyScanRequest::post( ':run_hardening' ) ) {
        if ( DesskyScanRequest::post( ':harden_readme' ) && $cp == 0 ) {
            if ( @unlink( ABSPATH.'/readme.html' ) === false ) {
                $upmsg = DesskyScanInterface::error( 'Unable to remove <code>readme.html</code> file.' );
            } else {
                $cp = 1;
                $message = 'Hardening applied to the <code>readme.html</code> file';
                DesskyScanEvent::report_notice_event( $message );
                DesskyScanInterface::info( $message );
            }
        } elseif ( DesskyScanRequest::post( ':harden_readme_unharden' ) ) {
            DesskyScanInterface::error( 'We can not revert this action, you must create the <code>readme.html</code> manually.' );
        }
    }

    return desskyscan_harden_status(
        'Information leakage (readme.html)',
        $cp,
        ( $cp == 0 ? 'desskyscan_harden_readme' : null ),
        '<code>readme.html</code> file properly deleted',
        '<code>readme.html</code> not deleted and leaking the WordPress version',
        'It checks whether you have the <code>readme.html</code> file available that leaks your WordPress version',
        $upmsg
    );
}

/**
 * Check whether the main administrator user still has the default name "admin"
 * or not, which can lead to an attacker to perform a brute force attack.
 *
 * @return void
 */
function desskyscan_harden_adminuser(){
    global $wpdb;

    $upmsg = null;
    $user_query = new WP_User_Query(array(
        'search' => 'admin',
        'fields' => array( 'ID', 'user_login' ),
        'search_columns' => array( 'user_login' ),
    ));
    $results = $user_query->get_results();
    $account_removed = ( count( $results ) === 0 ? 1 : 0 );

    return desskyscan_harden_status(
        'Default admin account',
        $account_removed,
        null,
        'Default admin user account (admin) not being used',
        'Default admin user account (admin) being used. Not recommended',
        'It checks whether you have the default <code>admin</code> account enabled, security guidelines recommend creating a new admin user name.',
        'admin username could be easily changed in the wordpress itself on the <a href="'.admin_url( 'users.php' ).'" target="_blank">user list</a> page by creating a new username with administrator role and by deleting old admin account afterwards.<br /><strong>While deleting your old user, WordPress will ask what you want to do with their content. Make sure that you click on ‘Attribute all content to:’ option and then select the new user you just created. All content from a previous user that is not attributed to a new user will be deleted! </strong>',
        2
    );
}

/**
 * Enable or disable the user of the built-in Wordpress file editor.
 *
 * @return void
 */
function desskyscan_harden_fileeditor(){
    $file_editor_disabled = defined( 'DISALLOW_FILE_EDIT' ) ? DISALLOW_FILE_EDIT : false;

    if ( DesskyScanRequest::post( ':run_hardening' ) ) {
        $current_time = date( 'r' );
        $wp_config_path = DesskyScan::get_wpconfig_path();

        $wp_config_writable = ( file_exists( $wp_config_path ) && is_writable( $wp_config_path ) ) ? true : false;
        $new_wpconfig = $wp_config_writable ? @file_get_contents( $wp_config_path ) : '';

        if ( DesskyScanRequest::post( ':harden_fileeditor' ) ) {
            if ( $wp_config_writable ) {
                if ( preg_match( '/(.*define\(.DB_COLLATE..*)/', $new_wpconfig, $match ) ) {
                    $disallow_fileedit_definition = "\n\ndefine('DISALLOW_FILE_EDIT', TRUE); // Dessky Security: {$current_time}\n";
                    $new_wpconfig = str_replace( $match[0], $match[0].$disallow_fileedit_definition, $new_wpconfig );
                }

                $file_editor_disabled = true;
                @file_put_contents( $wp_config_path, $new_wpconfig, LOCK_EX );
                $message = 'Hardening applied to the plugin and theme editor';
                DesskyScanEvent::report_notice_event( $message );
                DesskyScanInterface::info( $message );
            } else {
                DesskyScanInterface::error( 'The <code>wp-config.php</code> file is not in the default location
                    or is not writable, you will need to put the following code manually there:
                    <code>define("DISALLOW_FILE_EDIT", TRUE);</code>' );
            }
        } elseif ( DesskyScanRequest::post( ':harden_fileeditor_unharden' ) ) {
            if ( preg_match( "/(.*define\('DISALLOW_FILE_EDIT', TRUE\);.*)/", $new_wpconfig, $match ) ) {
                if ( $wp_config_writable ) {
                    $new_wpconfig = str_replace( "\n{$match[1]}", '', $new_wpconfig );
                    file_put_contents( $wp_config_path, $new_wpconfig, LOCK_EX );
                    $file_editor_disabled = false;
                    $message = 'Hardening reverted in the plugin and theme editor';
                    DesskyScanEvent::report_error_event( $message );
                    DesskyScanInterface::info( $message );
                } else {
                    DesskyScanInterface::error( 'The <code>wp-config.php</code> file is not in the default location
                        or is not writable, you will need to remove the following code manually from there:
                        <code>define("DISALLOW_FILE_EDIT", TRUE);</code>' );
                }
            } else {
                DesskyScanInterface::error( 'The theme and plugin editor are not disabled from the configuration file.' );
            }
        }
    }

    $message = 'Occasionally you may wish to disable the plugin or theme editor to prevent overzealous
        users from being able to edit sensitive files and potentially crash the site. Disabling these
        also provides an additional layer of security if a hacker gains access to a well-privileged
        user account.';

    return desskyscan_harden_status(
        'Plugin &amp; Theme editor',
        ( $file_editor_disabled === false ? 0 : 1 ),
        'desskyscan_harden_fileeditor',
        'File editor for Plugins and Themes is disabled',
        'File editor for Plugins and Themes is enabled',
        $message,
        null
    );
}

/**
 * Check whether the prefix of each table in the database designated for the site
 * is the same as the default prefix defined by Wordpress "_wp", in that case the
 * "harden" button will generate randomly a new prefix and rename all those tables.
 *
 * @return void
 */
function desskyscan_harden_dbtables(){
    global $table_prefix;

    $hardened = ( $table_prefix == 'wp_' ? 0 : 1 );

    return desskyscan_harden_status(
        'Database table prefix',
        $hardened,
        null,
        'Database table prefix properly modified',
        'Database table set to the default value <code>wp_</code>.',
        'It checks whether your database table prefix has been changed from the default <code>wp_</code>',
        'In most cases you will be fine with the default <code>wp_</code> prefix.<br /><strong>Bare in mind that this change is little tricky and if you are unfamiliar with databases and backups we suggest that you hire a professional developer to help you with the database prefix change.</strong>',
        1
    );
}

function desskyscan_harden_nginx_phpfpm(){
    $description = 'It seems that you are using the Nginx web server, if that is
        the case then you will need to add the following code into the global
        <code>nginx.conf</code> file or the virtualhost associated with this
        website. Choose the correct rules for the directories that you want to
        protect. If you encounter errors after restart the web server then revert
        the changes and contact the support team of your hosting company, or read
        the official article about <a href="http://codex.wordpress.org/Nginx">
        WordPress on Nginx</a>.</p>';

    $description .= "<pre class='code'># Block PHP files in uploads directory.\nlocation ~* /(?:uploads|files)/.*\.php$ {\n\x20\x20deny all;\n}</pre>";
    $description .= "<pre class='code'># Block PHP files in content directory.\nlocation ~* /wp-content/.*\.php$ {\n\x20\x20deny all;\n}</pre>";
    $description .= "<pre class='code'># Block PHP files in includes directory.\nlocation ~* /wp-includes/.*\.php$ {\n\x20\x20deny all;\n}</pre>";

    $description .= "<pre class='code'>";
    $description .= "# Block PHP files in uploads, content, and includes directory.\n";
    $description .= "location ~* /(?:uploads|files|wp-content|wp-includes)/.*\.php$ {\n";
    $description .= "\x20\x20deny all;\n";
    $description .= '}</pre>';

    $description .= '<p class="desskyscan-hidden">';

    return desskyscan_harden_status(
        'Block PHP files',
        999,
        null,
        null,
        null,
        $description
    );
}

/**
 * Check whether the WordPress upload folder is protected or not.
 *
 * A htaccess file is placed in the upload folder denying the access to any php
 * file that could be uploaded through a vulnerability in a Plugin, Theme or
 * WordPress itself.
 *
 * @return void
 */
function desskyscan_harden_upload(){
    $dpath = WP_CONTENT_DIR . '/uploads';

    if ( DesskyScanRequest::post( ':run_hardening' ) ) {
        if ( DesskyScanRequest::post( ':harden_upload' ) ) {
            $result = DesskyScanHardening::harden_directory( $dpath );

            if ( $result === true ) {
                $message = 'Hardening applied to the uploads directory';
                DesskyScanEvent::report_notice_event( $message );
                DesskyScanInterface::info( $message );
            } else {
                DesskyScanInterface::error( 'Error hardening directory, check the permissions.' );
            }
        } elseif ( DesskyScanRequest::post( ':harden_upload_unharden' ) ) {
            $result = DesskyScanHardening::unharden_directory( $dpath );

            if ( $result === true ) {
                $message = 'Hardening reverted in the uploads directory';
                DesskyScanEvent::report_error_event( $message );
                DesskyScanInterface::info( $message );
            } else {
                DesskyScanInterface::info( 'Access file is not writable, check the permissions.' );
            }
        }
    }

    // Check whether the directory is already hardened or not.
    $is_hardened = DesskyScanHardening::is_hardened( $dpath );
    $cp = ( $is_hardened === true ) ? 1 : 0;

    $description = 'It checks if the uploads directory of this site allows the direct execution'
        . ' of PHP files. It is recommendable to prevent this because someone may try to exploit'
        . ' a vulnerability of a plugin, theme, and/or other PHP-based code located in this'
        . ' directory sending requests directory to these files.</p><p><b>Note:</b> Many plugins'
        . ' and themes in the WordPress marketplace put <em>(insecure)</em> PHP files in this'
        . ' folder for <em>"X"</em> or <em>"Y"</em> reasons, they may not want to change their'
        . ' code to prevent security issues, so you will have to keep this option un-hardened'
        . ' or else you will end up breaking their functionality.';

    return desskyscan_harden_status(
        'Protect uploads directory',
        $cp,
        'desskyscan_harden_upload',
        'Upload directory properly hardened',
        'Upload directory not hardened',
        $description,
        null
    );
}

/**
 * Check whether the WordPress content folder is protected or not.
 *
 * A htaccess file is placed in the content folder denying the access to any php
 * file that could be uploaded through a vulnerability in a Plugin, Theme or
 * WordPress itself.
 *
 * @return void
 */
function desskyscan_harden_wpcontent(){
    if ( DesskyScanRequest::post( ':run_hardening' ) ) {
        if ( DesskyScanRequest::post( ':harden_wpcontent' ) ) {
            $result = DesskyScanHardening::harden_directory( WP_CONTENT_DIR );

            if ( $result === true ) {
                $message = 'Hardening applied to the content directory';
                DesskyScanEvent::report_notice_event( $message );
                DesskyScanInterface::info( $message );
            } else {
                DesskyScanInterface::error( 'Error hardening directory, check the permissions.' );
            }
        } elseif ( DesskyScanRequest::post( ':harden_wpcontent_unharden' ) ) {
            $result = DesskyScanHardening::unharden_directory( WP_CONTENT_DIR );

            if ( $result === true ) {
                $message = 'Hardening reverted in the content directory';
                DesskyScanEvent::report_error_event( $message );
                DesskyScanInterface::info( $message );
            } else {
                DesskyScanInterface::info( 'Access file is not writable, check the permissions.' );
            }
        }
    }

    // Check whether the directory is already hardened or not.
    $is_hardened = DesskyScanHardening::is_hardened( WP_CONTENT_DIR );
    $cp = ( $is_hardened === true ) ? 1 : 0;

    $description = 'This option blocks direct access to any PHP file located under the content'
        . ' directory of this site. The note under the <em>"Protect uploads directory"</em>'
        . ' section also applies to this option so you may want to read that part too. If you'
        . ' experience any kind of issues in your site after you apply this hardening go to the'
        . ' content directory using a FTP client or a file manager <em>(generally available in'
        . ' your hosting panel)</em> and rename a file named <code>.htaccess</code>.';

    return desskyscan_harden_status(
        'Restrict wp-content access',
        $cp,
        'desskyscan_harden_wpcontent',
        'WP-content directory properly hardened',
        'WP-content directory not hardened',
        $description,
        null
    );
}

/**
 * Check whether the WordPress includes folder is protected or not.
 *
 * A htaccess file is placed in the includes folder denying the access to any php
 * file that could be uploaded through a vulnerability in a Plugin, Theme or
 * WordPress itself, there are some exceptions for some specific files that must
 * be available publicly.
 *
 * @return void
 */
function desskyscan_harden_wpincludes(){
    $dpath = ABSPATH . '/wp-includes';

    if ( DesskyScanRequest::post( ':run_hardening' ) ) {
        if ( DesskyScanRequest::post( ':harden_wpincludes' ) ) {
            $result = DesskyScanHardening::harden_directory( $dpath );

            if ( $result === true ) {
                $message = 'Hardening applied to the library directory';
                DesskyScanEvent::report_notice_event( $message );
                DesskyScanInterface::info( $message );
            } else {
                DesskyScanInterface::error( 'Error hardening directory, check the permissions.' );
            }
        } elseif ( DesskyScanRequest::post( ':harden_wpincludes_unharden' ) ) {
            $result = DesskyScanHardening::unharden_directory( $dpath );

            if ( $result === true ) {
                $message = 'Hardening reverted in the library directory';
                DesskyScanEvent::report_error_event( $message );
                DesskyScanInterface::info( $message );
            } else {
                DesskyScanInterface::info( 'Access file is not writable, check the permissions.' );
            }
        }
    }

    // Check whether the directory is already hardened or not.
    $is_hardened = DesskyScanHardening::is_hardened( $dpath );
    $cp = ( $is_hardened === true ) ? 1 : 0;

    return desskyscan_harden_status(
        'Restrict wp-includes access',
        $cp,
        'desskyscan_harden_wpincludes',
        'WP-Includes directory properly hardened',
        'WP-Includes directory not hardened',
        'This option blocks direct PHP access to any file inside <code>wp-includes</code>.',
        null
    );
}

/**
 * Generate the HTML code necessary to show a form with the options to harden
 * a specific part of the WordPress installation, if the Status variable is
 * set as a positive integer the button is shown as "unharden".
 *
 * @param  string  $title       Title of the panel.
 * @param  integer $status      Either one or zero representing the state of the hardening, one for secure, zero for insecure.
 * @param  string  $type        Name of the hardening option, this will be used through out the form generation.
 * @param  string  $messageok   Message that will be shown if the hardening was executed.
 * @param  string  $messagewarn Message that will be shown if the hardening is not executed.
 * @param  string  $desc        Optional description of the hardening.
 * @param  string  $updatemsg   Optional explanation of the hardening after the submission of the form.
 * @param  integer $optionaltpl Use alternative template file
 * @return void
 */
function desskyscan_harden_status( $title = '', $status = 0, $type = '', $messageok = '', $messagewarn = '', $desc = null, $updatemsg = null, $optionaltpl = 0 ){
    $template_variables = array(
        'Hardening.Title' => DesskyScan::escape( $title ),
        'Hardening.Description' => '',
        'Hardening.Status' => 'unknown',
        'Hardening.FieldName' => '',
        'Hardening.FieldValue' => '',
        'Hardening.FieldAttributes' => '',
        'Hardening.Information' => '',
        'Hardening.UpdateMessage' => '',
    );

    if ( is_null( $type ) ) {
        $type = 'unknown';
        $template_variables['Hardening.FieldAttributes'] = 'disabled="disabled"';
    }

    $template_variables['Hardening.Status'] = (string) $status;

    if ( $status === 1 ) {
        $template_variables['Hardening.FieldName'] = $type . '_unharden';
        $template_variables['Hardening.FieldValue'] = 'Revert hardening';
        $template_variables['Hardening.Information'] = $messageok;
    } elseif ( $status === 0 ) {
        $template_variables['Hardening.FieldName'] = $type;
        $template_variables['Hardening.FieldValue'] = 'Harden';
        $template_variables['Hardening.Information'] = $messagewarn;
    } else {
        $template_variables['Hardening.FieldName'] = '';
        $template_variables['Hardening.FieldValue'] = 'Unavailable';
        $template_variables['Hardening.Information'] = 'Can not be determined.';
        $template_variables['Hardening.FieldAttributes'] = 'disabled="disabled"';
    }

    if ( ! is_null( $desc ) ) {
        $template_variables['Hardening.Description'] = '<p>' . $desc . '</p>';
    }

    if ( ! is_null( $updatemsg ) ) {
        $template_variables['Hardening.UpdateMessage'] = '<p>' . $updatemsg . '</p>';
    }
	
	if ( $optionaltpl === 1 ) {
		$optionaltpl = '-optional';
	}elseif ( $optionaltpl === 2 ) {
		$optionaltpl = '-info';
	}else{
		$optionaltpl = '';
	}

    return DesskyScanTemplate::get_snippet( 'hardening'.$optionaltpl, $template_variables );
}

/**
 * Generate and print the HTML code for the InfoSys page.
 *
 * This page will contains information of the system where the site is hosted,
 * also information about users in session, htaccess rules and configuration
 * options.
 *
 * @return void
 */
function desskyscan_infosys_page(){
    DesskyScanInterface::check_permissions();

    // Process all form submissions.
    desskyscan_infosys_form_submissions();

    // Page pseudo-variables initialization.
    $template_variables = array(
        'PageTitle' => 'Site Info',
        'ServerInfo' => desskyscan_server_info(),
        'HTAccessIntegrity' => desskyscan_infosys_htaccess(),
        'WordpressConfig' => desskyscan_infosys_wpconfig(),
        'ErrorLogs' => desskyscan_infosys_errorlogs(),
    );

    echo DesskyScanTemplate::get_template( 'infosys', $template_variables );
}


/**
 * Locate, parse and display the latest error logged in the main error_log file.
 *
 * @return array A list of pseudo-variables and values that will replace them in the HTML template.
 */
function desskyscan_infosys_errorlogs(){
    $template_variables = array(
        'ErrorLog.Path' => '',
        'ErrorLog.Exists' => 'No',
        'ErrorLog.NoItemsVisibility' => 'hidden',
        'ErrorLog.DisabledVisibility' => 'hidden',
        'ErrorLog.InvalidFormatVisibility' => 'hidden',
        'ErrorLog.LogsLimit' => '0',
        'ErrorLog.FileSize' => '0B',
        'ErrorLog.List' => '',
    );

    $error_log_path = false;
    $log_filename = DesskyScan::ini_get( 'error_log' );
    $errorlogs_limit = DesskyScanOption::get_option( ':errorlogs_limit' );
    $template_variables['ErrorLog.LogsLimit'] = $errorlogs_limit;
    $errorlogs_counter = 0;

    if ( $log_filename ) {
        $error_log_path = @realpath( ABSPATH . '/' . $log_filename );
    }

    if ( DesskyScanOption::get_option( ':parse_errorlogs' ) === 'disabled' ) {
        $template_variables['ErrorLog.DisabledVisibility'] = 'visible';
    }

    if ( $error_log_path ) {
        $template_variables['ErrorLog.Path'] = $error_log_path;
        $template_variables['ErrorLog.Exists'] = 'Yes';
        $template_variables['ErrorLog.FileSize'] = DesskyScan::human_filesize( filesize( $error_log_path ) );

        $last_lines = DesskyScanFileInfo::tail_file( $error_log_path, $errorlogs_limit );
        $error_logs = DesskyScanFSScanner::parse_error_logs( $last_lines );
        $error_logs = array_reverse( $error_logs );
        $errorlogs_counter = 0;

        foreach ( $error_logs as $error_log ) {
            $css_class = ( $errorlogs_counter % 2 == 0 ) ? '' : 'alternate';
            $template_variables['ErrorLog.List'] .= DesskyScanTemplate::get_snippet('infosys-errorlogs', array(
                'ErrorLog.CssClass' => $css_class,
                'ErrorLog.DateTime' => DesskyScan::datetime( $error_log->timestamp ),
                'ErrorLog.ErrorType' => DesskyScan::escape( $error_log->error_type ),
                'ErrorLog.ErrorCode' => DesskyScan::escape( $error_log->error_code ),
                'ErrorLog.ErrorAbbr' => strtoupper( substr( $error_log->error_code, 0, 1 ) ),
                'ErrorLog.ErrorMessage' => DesskyScan::escape( $error_log->error_message ),
                'ErrorLog.FilePath' => DesskyScan::escape( $error_log->file_path ),
                'ErrorLog.LineNumber' => DesskyScan::escape( $error_log->line_number ),
            ));
            $errorlogs_counter += 1;
        }

        if ( $errorlogs_counter <= 0 ) {
            $template_variables['ErrorLog.InvalidFormatVisibility'] = 'visible';
        }
    } else {
        $template_variables['ErrorLog.NoItemsVisibility'] = 'visible';
    }

    return DesskyScanTemplate::get_section( 'infosys-errorlogs', $template_variables );
}

/**
 * Find the main htaccess file for the site and check whether the rules of the
 * main htaccess file of the site are the default rules generated by WordPress.
 *
 * @return string The HTML code displaying the information about the HTAccess rules.
 */
function desskyscan_infosys_htaccess(){
    $htaccess_path = DesskyScan::get_htaccess_path();

    $template_variables = array(
        'HTAccess.Content' => '',
        'HTAccess.Message' => '',
        'HTAccess.MessageType' => '',
        'HTAccess.MessageVisible' => 'hidden',
        'HTAccess.TextareaVisible' => 'hidden',
    );

    if ( $htaccess_path ) {
        $htaccess_rules = file_get_contents( $htaccess_path );

        $template_variables['HTAccess.MessageType'] = 'updated';
        $template_variables['HTAccess.MessageVisible'] = 'visible';
        $template_variables['HTAccess.TextareaVisible'] = 'visible';
        $template_variables['HTAccess.Content'] = $htaccess_rules;
        $template_variables['HTAccess.Message'] .= 'HTAccess file found in this path <code>'.$htaccess_path.'</code>';

        if ( empty($htaccess_rules) ) {
            $template_variables['HTAccess.TextareaVisible'] = 'hidden';
            $template_variables['HTAccess.Message'] .= '</p><p>The HTAccess file found is completely empty.';
        }
        if ( desskyscan_htaccess_is_standard( $htaccess_rules ) ) {
            $template_variables['HTAccess.Message'] .= '</p><p>
                The main <code>.htaccess</code> file in your site has the standard rules for a WordPress installation. You can customize it to improve the
                performance and change the behaviour of the redirections for pages and posts in your site. To get more information visit the official documentation at
                <a href="http://codex.wordpress.org/Using_Permalinks#Creating_and_editing_.28.htaccess.29" target="_blank">Codex WordPrexx - Creating and editing (.htaccess)</a>';
        }
    } else {
        $template_variables['HTAccess.Message'] = 'Your website does not contains a <code>.htaccess</code> file or it was not found in the default location.';
        $template_variables['HTAccess.MessageType'] = 'error';
        $template_variables['HTAccess.MessageVisible'] = 'visible';
    }

    return DesskyScanTemplate::get_section( 'infosys-htaccess', $template_variables );
}

/**
 * Retrieve all the constants and variables with their respective values defined
 * in the WordPress configuration file, only the database password constant is
 * omitted for security reasons.
 *
 * @return string The HTML code displaying the constants and variables found in the wp-config file.
 */
function desskyscan_infosys_wpconfig(){
    $template_variables = array(
        'WordpressConfig.Rules' => '',
        'WordpressConfig.Total' => 0,
    );

    $ignore_wp_rules = array( 'DB_PASSWORD' );
    $wp_config_path = DesskyScan::get_wpconfig_path();

    if ( $wp_config_path ) {
        $wp_config_rules = array();
        $wp_config_content = DesskyScanFileInfo::file_lines( $wp_config_path );

        // Parse the main configuration file and look for constants and global variables.
        foreach ( (array) $wp_config_content as $line ) {
            if ( preg_match( '/^\s?(#|\/\/)/', $line ) ) {
                // Ignore commented lines.
                continue;
            } elseif ( preg_match( '/define\(/', $line ) ) {
                // Detect PHP constants even if the line if indented.
                $line = preg_replace( '/.*define\((.+)\);.*/', '$1', $line );
                $line_parts = explode( ',', $line, 2 );
            } elseif ( preg_match( '/^\$[a-zA-Z_]+/', $line ) ) {
                // Detect global variables like the database table prefix.
                $line = preg_replace( '/;\s\/\/.*/', ';', $line );
                $line_parts = explode( '=', $line, 2 );
            } else {
                // Ignore other lines.
                continue;
            }

            // Clean and append the rule to the wp_config_rules variable.
            if ( isset($line_parts) && count( $line_parts ) == 2 ) {
                $key_name = '';
                $key_value = '';

                // TODO: A foreach loop is not really necessary, find a better way.
                foreach ( $line_parts as $i => $line_part ) {
                    $line_part = trim( $line_part );
                    $line_part = ltrim( $line_part, '$' );
                    $line_part = rtrim( $line_part, ';' );

                    // Remove single/double quotes at the beginning and end of the string.
                    $line_part = ltrim( $line_part, "'" );
                    $line_part = rtrim( $line_part, "'" );
                    $line_part = ltrim( $line_part, '"' );
                    $line_part = rtrim( $line_part, '"' );

                    // Assign the clean strings to specific variables.
                    if ( $i == 0 ) {
                        $key_name = $line_part;
                    }

                    if ( $i == 1 ) {
                        if ( defined( $key_name ) ) {
                            $key_value = constant( $key_name );

                            if ( is_bool( $key_value ) ) {
                                $key_value = ( $key_value === true ) ? 'TRUE' : 'FALSE';
                            }
                        } else {
                            $key_value = $line_part;
                        }
                    }
                }

                // Remove the value of sensitive variables like the database password.
                if ( in_array( $key_name, $ignore_wp_rules ) ) {
                    $key_value = 'hidden';
                }

                // Append the value to the configuration rules.
                $wp_config_rules[ $key_name ] = $key_value;
            }
        }

        // Pass the WordPress configuration rules to the template and show them.
        $counter = 0;
        foreach ( $wp_config_rules as $var_name => $var_value ) {
            $css_class = ( $counter % 2 == 0 ) ? '' : 'alternate';
            $label_css = 'desskyscan-monospace';

            if ( empty($var_value) ) {
                $var_value = 'empty';
                $label_css = 'desskyscan-label-default';
            } elseif ( $var_value == 'hidden' ) {
                $label_css = 'desskyscan-label-info';
            }

            $template_variables['WordpressConfig.Total'] += 1;
            $template_variables['WordpressConfig.Rules'] .= DesskyScanTemplate::get_snippet('infosys-wpconfig', array(
                'WordpressConfig.VariableName' => DesskyScan::escape( $var_name ),
                'WordpressConfig.VariableValue' => DesskyScan::escape( $var_value ),
                'WordpressConfig.VariableCssClass' => $label_css,
                'WordpressConfig.CssClass' => $css_class,
            ));
            $counter += 1;
        }
    }

    return DesskyScanTemplate::get_section( 'infosys-wpconfig', $template_variables );
}

/**
 * Process the requests sent by the form submissions originated in the infosys
 * page, all forms must have a nonce field that will be checked against the one
 * generated in the template render function.
 *
 * @param  boolean $page_nonce True if the nonce is valid, False otherwise.
 * @return void
 */
function desskyscan_infosys_form_submissions(){
    if ( DesskyScanInterface::check_nonce() ) {
        // Modify the scheduled tasks (run now, remove, re-schedule).
        $allowed_actions = '(runnow|hourly|twicedaily|daily|remove)';

        if ( $cronjob_action = DesskyScanRequest::post( ':cronjob_action', $allowed_actions ) ) {
            $cronjobs = DesskyScanRequest::post( ':cronjobs', '_array' );

            if ( ! empty($cronjobs) ) {
                $total_tasks = count( $cronjobs );

                // Force execution of the selected scheduled tasks.
                if ( $cronjob_action == 'runnow' ) {
                    DesskyScanInterface::info( $total_tasks . ' tasks were scheduled to run in the next ten seconds.' );
                    DesskyScanEvent::report_notice_event( sprintf(
                        'Force execution of scheduled tasks: (multiple entries): %s',
                        @implode( ',', $cronjobs )
                    ) );

                    foreach ( $cronjobs as $task_name ) {
                        wp_schedule_single_event( time() + 10, $task_name );
                    }
                }

                // Force deletion of the selected scheduled tasks.
                elseif ( $cronjob_action == 'remove' ) {
                    DesskyScanInterface::info( $total_tasks . ' scheduled tasks were removed.' );
                    DesskyScanEvent::report_notice_event( sprintf(
                        'Delete scheduled tasks: (multiple entries): %s',
                        @implode( ',', $cronjobs )
                    ) );

                    foreach ( $cronjobs as $task_name ) {
                        wp_clear_scheduled_hook( $task_name );
                    }
                }

                // Re-schedule the selected scheduled tasks.
                elseif (
                    $cronjob_action == 'hourly'
                    || $cronjob_action == 'twicedaily'
                    || $cronjob_action == 'daily'
                ) {
                    DesskyScanInterface::info( $total_tasks . ' tasks were re-scheduled to run <code>' . $cronjob_action . '</code>.' );
                    DesskyScanEvent::report_notice_event( sprintf(
                        'Re-configure scheduled tasks %s: (multiple entries): %s',
                        $cronjob_action,
                        @implode( ',', $cronjobs )
                    ) );

                    foreach ( $cronjobs as $task_name ) {
                        wp_clear_scheduled_hook( $task_name );
                        $next_due = wp_next_scheduled( $task_name );
                        wp_schedule_event( $next_due, $cronjob_action, $task_name );
                    }
                }
            } else {
                DesskyScanInterface::error( 'No scheduled tasks were selected from the list.' );
            }
        }
    }
}

/**
 * Gather information from the server, database engine, and PHP interpreter.
 *
 * @return array A list of pseudo-variables and values that will replace them in the HTML template.
 */
function desskyscan_server_info(){
    global $wpdb;

    $template_variables = array(
        'ServerInfo.Variables' => '',
    );

    $info_vars = array(
        'Plugin_version' => DESSKYSCAN_VERSION,
        'Plugin_checksum' => DESSKYSCAN_PLUGIN_CHECKSUM,
        'Host_Address' => 'Unknown',
        'HTTP_Host' => 'Unknown',
        'Host_Name' => 'Unknown',
        'Site_URL' => 'Unknown',
        'Top_Level_Domain' => 'Unknown',
        'Remote_Address' => DesskyScan::get_remote_addr(),
        'Remote_Address_Header' => DesskyScan::get_remote_addr_header(),
        'Operating_system' => sprintf( '%s (%d Bit)', PHP_OS, PHP_INT_SIZE * 8 ),
        'Server' => 'Unknown',
        'Developer_mode' => 'OFF',
        'Memory_usage' => 'N/A',
        'MySQL_version' => '0.0',
        'SQL_mode' => 'Not set',
        'PHP_version' => PHP_VERSION,
    );

    $proxy_info = DesskyScan::check_hosting_server( true );

    $info_vars['HTTP_Host'] = $proxy_info['http_host'];
    $info_vars['Host_Name'] = $proxy_info['host_name'];
    $info_vars['Host_Address'] = $proxy_info['host_addr'];
    $info_vars['Site_URL'] = DesskyScan::get_domain();
    $info_vars['Top_Level_Domain'] = DesskyScan::get_domain( true );

    if ( defined( 'WP_DEBUG' ) && WP_DEBUG ) {
        $info_vars['Developer_mode'] = 'ON';
    }

    if ( function_exists( 'memory_get_usage' ) ) {
        $info_vars['Memory_usage'] = round( memory_get_usage() / 1024 / 1024, 2 ).' MB';
    }

    if ( isset($_SERVER['SERVER_SOFTWARE']) ) {
        $info_vars['Server'] = DesskyScan::escape( $_SERVER['SERVER_SOFTWARE'] );
    }

    if ( $wpdb ) {
        $info_vars['MySQL_version'] = $wpdb->get_var( 'SELECT VERSION() AS version' );

        $mysql_info = $wpdb->get_results( 'SHOW VARIABLES LIKE "sql_mode"' );
        if ( is_array( $mysql_info ) && ! empty($mysql_info[0]->Value) ) {
            $info_vars['SQL_mode'] = $mysql_info[0]->Value;
        }
    }

    $field_names = array(
        'safe_mode',
        'expose_php',
        'allow_url_fopen',
        'memory_limit',
        'upload_max_filesize',
        'post_max_size',
        'max_execution_time',
        'max_input_time',
    );

    foreach ( $field_names as $php_flag ) {
        $php_flag_value = DesskyScan::ini_get( $php_flag );
        $php_flag_name = 'PHP_' . $php_flag;
        $info_vars[ $php_flag_name ] = $php_flag_value ? $php_flag_value : 'N/A';
    }

    $counter = 0;

    foreach ( $info_vars as $var_name => $var_value ) {
        $css_class = ( $counter % 2 == 0 ) ? '' : 'alternate';
        $var_name = str_replace( '_', chr( 32 ), $var_name );

        $template_variables['ServerInfo.Variables'] .= DesskyScanTemplate::get_snippet('infosys-serverinfo', array(
            'ServerInfo.CssClass' => $css_class,
            'ServerInfo.Title' => $var_name,
            'ServerInfo.Value' => $var_value,
        ));
        $counter += 1;
    }

    return DesskyScanTemplate::get_section( 'infosys-serverinfo', $template_variables );
}

/**
 * Check whether the rules in a htaccess file are the default options generated
 * by WordPress or if the file has custom options added by other Plugins.
 *
 * @param  string  $rules Optional parameter containing a text string with the content of the main htaccess file.
 * @return boolean        Either TRUE or FALSE if the rules found in the htaccess file specified are the default ones or not.
 */
function desskyscan_htaccess_is_standard( $rules = false ){
    if ( $rules === false ) {
        $htaccess_path = DesskyScan::get_htaccess_path();
        $rules = $htaccess_path ? file_get_contents( $htaccess_path ) : '';
    }

    if ( ! empty($rules) ) {
        $standard_lines = array(
            '# BEGIN WordPress',
            '<IfModule mod_rewrite\.c>',
            'RewriteEngine On',
            'RewriteBase \/',
            'RewriteRule .index.\.php. - \[L\]',
            'RewriteCond %\{REQUEST_FILENAME\} \!-f',
            'RewriteCond %\{REQUEST_FILENAME\} \!-d',
            'RewriteRule \. \/index\.php \[L\]',
            '<\/IfModule>',
            '# END WordPress',
        );
        $pattern  = '';
        $standard_lines_total = count( $standard_lines );
        foreach ( $standard_lines as $i => $line ) {
            if ( $i < ($standard_lines_total -1) ) {
                $end_of_line = "\n";
            } else {
                $end_of_line = '';
            }
            $pattern .= sprintf( '%s%s', $line, $end_of_line );
        }

        if ( preg_match( "/{$pattern}/", $rules ) ) {
            return true;
        }
    }

    return false;
}